from baseTest import BaseTestCase
from jstreams import (
    stream,
    Not,
    equalsIgnoreCase,
    strContains,
    isNotNone,
    mapperOf,
    optional,
    predicateOf,
    reducerOf,
    toFloat,
)
from jstreams.predicate import hasKey, hasValue, isKeyIn, isValueIn
from jstreams.utils import identity


class TestPredicate(BaseTestCase):
    def test_predicate_and(self) -> None:
        expected = "Test"
        predicate = predicateOf(isNotNone).And(equalsIgnoreCase("Test"))
        self.assertEqual(
            optional("Test").filter(predicate).get(),
            expected,
            "Expected value should be correct",
        )

    def test_predicate_and2(self) -> None:
        expected = "test value"
        predicate = predicateOf(strContains("test")).And(strContains("value"))
        self.assertEqual(
            optional(expected).filter(predicate).get(),
            expected,
            "Expected value should be correct",
        )

    def test_predicate_and3(self) -> None:
        predicate = predicateOf(strContains("test")).And(Not(strContains("value")))
        self.assertListEqual(
            stream(["test value", "test other", "some value"])
            .filter(predicate)
            .toList(),
            ["test other"],
            "Expected value should be correct",
        )

    def test_predicate_or(self) -> None:
        predicate = predicateOf(strContains("es")).Or(equalsIgnoreCase("Other"))
        self.assertListEqual(
            stream(["Test", "Fest", "other", "Son", "Father"])
            .filter(predicate)
            .toList(),
            ["Test", "Fest", "other"],
            "Expected value should be correct",
        )

    def test_predicate_call(self) -> None:
        predicate = predicateOf(strContains("es"))
        self.assertTrue(
            predicate("test"),
            "Predicate should be callable and return the proper value",
        )

        self.assertTrue(
            predicate.Apply("test"),
            "Predicate should be callable via Apply and return the proper value",
        )
        self.assertFalse(
            predicate("nomatch"),
            "Predicate should be callable and return the proper value",
        )
        self.assertFalse(
            predicate.Apply("nomatch"),
            "Predicate should be callable via Apply and return the proper value",
        )

    def test_mapper_call(self) -> None:
        mapper = mapperOf(toFloat)
        self.assertEqual(
            mapper("1.2"), 1.2, "Mapper should be callable and return the proper value"
        )
        self.assertEqual(
            mapper.Map("1.2"),
            1.2,
            "Mapper should be callable via Map and return the proper value",
        )

    def test_reducer_call(self) -> None:
        reducer = reducerOf(max)
        self.assertEqual(
            reducer(1, 2), 2, "Reducer should be callable and return the proper value"
        )
        self.assertEqual(
            reducer.Reduce(1, 2),
            2,
            "Reducer should be callable via Reduce and return the proper value",
        )

    def test_dict_keys_values(self) -> None:
        dct = {"test": "A"}
        self.assertTrue(hasKey("test")(dct), "Dict should contain key")
        self.assertTrue(hasValue("A")(dct), "Dict should contain value")
        self.assertFalse(hasKey("other")(dct), "Dict should not contain key")
        self.assertFalse(hasValue("B")(dct), "Dict should not contain value")
        self.assertTrue(isKeyIn(dct)("test"), "Dict should contain key")
        self.assertTrue(isValueIn(dct)("A"), "Dict should contain value")
        self.assertFalse(isKeyIn(dct)("other"), "Dict should not contain key")
        self.assertFalse(isValueIn(dct)("B"), "Dict should not contain value")

    def test_identity(self) -> None:
        initial = ["1", "2"]
        self.assertListEqual(
            stream(initial).map(identity).toList(),
            initial,
            "Lists should match after identity mapping",
        )



from baseTest import BaseTestCase
from jstreams import Stream


class TestStream(BaseTestCase):
    def test_stream_map(self) -> None:
        """
        Test stream map function
        """
        self.assertEqual(
            Stream(["Test", "Best", "Lest"]).map(str.upper).toList(),
            ["TEST", "BEST", "LEST"],
        )

    def test_stream_filter(self) -> None:
        """
        Test stream filter function
        """
        self.assertEqual(
            Stream(["Test", "Best", "Lest"])
            .filter(lambda s: s.startswith("T"))
            .toList(),
            ["Test"],
        )
        self.assertFalse(
            Stream(["Test", "Best", "Lest"])
            .filter(lambda s: s.startswith("T"))
            .isEmpty()
        )
        self.assertTrue(
            Stream(["Test", "Best", "Lest"])
            .filter(lambda s: s.startswith("T"))
            .isNotEmpty()
        )

        self.assertEqual(
            Stream(["Test", "Best", "Lest"])
            .filter(lambda s: s.startswith("X"))
            .toList(),
            [],
        )

        self.assertTrue(
            Stream(["Test", "Best", "Lest"])
            .filter(lambda s: s.startswith("X"))
            .isEmpty()
        )

        self.assertFalse(
            Stream(["Test", "Best", "Lest"])
            .filter(lambda s: s.startswith("X"))
            .isNotEmpty()
        )

    def test_stream_anyMatch(self) -> None:
        """
        Test stream anyMatch function
        """
        self.assertFalse(
            Stream(["Test", "Best", "Lest"]).anyMatch(lambda s: s.startswith("X"))
        )

        self.assertTrue(
            Stream(["Test", "Best", "Lest"]).anyMatch(lambda s: s.startswith("T"))
        )

    def test_stream_allMatch(self) -> None:
        """
        Test stream allMatch function
        """
        self.assertTrue(
            Stream(["Test", "Best", "Lest"]).allMatch(lambda s: s.endswith("est"))
        )

        self.assertFalse(
            Stream(["Test", "Best", "Lest1"]).allMatch(lambda s: s.endswith("est"))
        )

    def test_stream_noneMatch(self) -> None:
        """
        Test stream noneMatch function
        """
        self.assertFalse(
            Stream(["Test", "Best", "Lest"]).noneMatch(lambda s: s.endswith("est"))
        )

        self.assertTrue(
            Stream(["Test", "Best", "Lest1"]).noneMatch(lambda s: s.endswith("xx"))
        )

    def test_stream_findFirst(self) -> None:
        """
        Test stream findFirst function
        """

        self.assertEqual(
            Stream(["Test", "Best", "Lest"])
            .findFirst(lambda s: s.startswith("L"))
            .getActual(),
            "Lest",
        )

    def test_stream_first(self) -> None:
        """
        Test stream first function
        """

        self.assertEqual(
            Stream(["Test", "Best", "Lest"]).first().getActual(),
            "Test",
        )

    def test_stream_cast(self) -> None:
        """
        Test stream cast function
        """

        self.assertEqual(
            Stream(["Test1", "Test2", 1, 2])
            .filter(lambda el: el == "Test1")
            .cast(str)
            .first()
            .getActual(),
            "Test1",
        )

    def test_stream_flatMap(self) -> None:
        """
        Test stream flatMap function
        """

        self.assertEqual(
            Stream([["a", "b"], ["c", "d"]]).flatMap(list).toList(),
            ["a", "b", "c", "d"],
        )

    def test_stream_skip(self) -> None:
        """
        Test stream skip function
        """

        self.assertEqual(
            Stream(["a", "b", "c", "d"]).skip(2).toList(),
            ["c", "d"],
        )

    def test_stream_limit(self) -> None:
        """
        Test stream limit function
        """

        self.assertEqual(
            Stream(["a", "b", "c", "d"]).limit(2).toList(),
            ["a", "b"],
        )

    def test_stream_takeWhile(self) -> None:
        """
        Test stream takeWhile function
        """

        self.assertEqual(
            Stream(["a1", "a2", "a3", "b", "c", "d"])
            .takeWhile(lambda e: e.startswith("a"))
            .toList(),
            ["a1", "a2", "a3"],
        )

    def test_stream_reduce(self) -> None:
        """
        Test stream reduce function
        """

        self.assertEqual(
            Stream(["aaa", "aa", "aaaa", "b", "c", "d"])
            .reduce(lambda el1, el2: el1 if len(el1) > len(el2) else el2)
            .getActual(),
            "aaaa",
        )

    def test_stream_reduce_integers(self) -> None:
        """
        Test stream reduce function
        """

        self.assertEqual(
            Stream([1, 2, 3, 4, 20, 5, 6]).reduce(max).getActual(),
            20,
        )

    def test_stream_nonNull(self) -> None:
        """
        Test stream nonNull function
        """

        self.assertEqual(
            Stream(["A", None, "B", None, None, "C", None, None]).nonNull().toList(),
            ["A", "B", "C"],
        )

    def str_len_cmp(self, a: str, b: str) -> int:
        return len(b) - len(a)

    def test_stream_sort(self) -> None:
        """
        Test stream sort function
        """

        self.assertEqual(
            Stream(["1", "333", "22", "4444", "55555"]).sort(self.str_len_cmp).toList(),
            ["55555", "4444", "333", "22", "1"],
        )

    def test_stream_reverse(self) -> None:
        """
        Test stream reverse function
        """

        self.assertEqual(
            Stream(["1", "333", "22", "4444", "55555"])
            .sort(self.str_len_cmp)
            .reverse()
            .toList(),
            ["1", "22", "333", "4444", "55555"],
        )

    def test_stream_distinct(self) -> None:
        """
        Test stream distinct function
        """

        self.assertEqual(
            Stream(["1", "1", "2", "3", "3", "4"]).distinct().toList(),
            ["1", "2", "3", "4"],
        )

    def test_stream_dropWhile(self) -> None:
        """
        Test stream dropWhile function
        """

        self.assertEqual(
            Stream(["a1", "a2", "a3", "b", "c", "d"])
            .dropWhile(lambda e: e.startswith("a"))
            .toList(),
            ["b", "c", "d"],
        )

        self.assertEqual(
            Stream(["a1", "a2", "a3", "a4", "a5", "a6"])
            .dropWhile(lambda e: e.startswith("a"))
            .toList(),
            [],
        )

    def test_stream_concat(self) -> None:
        """
        Test stream concat function
        """

        self.assertEqual(
            Stream(["a", "b", "c", "d"]).concat(Stream(["e", "f"])).toList(),
            ["a", "b", "c", "d", "e", "f"],
        )

    def test_stream_flatten(self) -> None:
        """
        Test stream flattening
        """

        self.assertEqual(
            Stream([["A", "B"], ["C", "D"], ["E", "F"]]).flatten(str).toList(),
            ["A", "B", "C", "D", "E", "F"],
        )

        self.assertEqual(
            Stream(["A", "B"]).flatten(str).toList(),
            ["A", "B"],
        )



import json
from typing import Any, Callable, Optional, TypeVar, Union

T = TypeVar("T")


def _f() -> None:
    pass


class _F:
    def mth(self) -> None:
        pass


FnType = type(_f)
MthType = type(_F().mth)


def isCallable(var: Any) -> bool:
    """
    Checks if the given argument is either a function or a method in a class.

    Args:
        var (Any): The argument to check

    Returns:
        bool: True if var is a function or method, False otherwise
    """
    varType = type(var)
    return varType is FnType or varType is MthType


def requireNotNull(obj: Optional[T]) -> T:
    """
    Returns a non null value of the object provided. If the provided value is null,
    the function raises a ValueError.

    Args:
        obj (Optional[T]): The object

    Raises:
        ValueError: Thrown when obj is None

    Returns:
        T: The non null value
    """
    if obj is None:
        raise ValueError("None object provided")
    return obj


def isNumber(anyVal: Any) -> bool:
    """Checks if the value provided is a number

    Args:
        anyVal (any): the value

    Returns:
        bool: True if anyVal is a number, False otherwise
    """
    try:
        _: float = float(anyVal) + 1
    except Exception:
        return False
    return True


def toInt(val: Any) -> int:
    """
    Returns an int representation of the given value.
    Raises a ValueError if the value cannot be represented as an int.

    Args:
        val (Any): The value

    Returns:
        int: The int representation
    """
    return int(str(val))


def toFloat(val: Any) -> float:
    """
    Returns a float representation of the given value.
    Raises a ValueError if the value cannot be represented as a float.

    Args:
        val (Any): The value

    Returns:
        float: The float representation
    """
    return float(str(val))


def asList(dct: dict[Any, T]) -> list[T]:
    """
    Returns the values in a dict as a list.

    Args:
        dct (dict[Any, T]): The dictionary

    Returns:
        list[T]: The list of values
    """
    return [v for _, v in dct.items()]


def keysAsList(dct: dict[T, Any]) -> list[T]:
    """
    Returns the keys in a dict as a list

    Args:
        dct (dict[T, Any]): The dictionary

    Returns:
        list[T]: The list of keys
    """
    return [k for k, _ in dct.items()]


def loadJson(
    s: Union[str, bytes, bytearray],
) -> Optional[Union[list[Any], dict[Any, Any]]]:
    return loadJsonEx(s, None)


def loadJsonEx(
    s: Union[str, bytes, bytearray], handler: Optional[Callable[[Exception], Any]]
) -> Optional[Union[list[Any], dict[Any, Any]]]:
    try:
        return json.loads(s)  # type: ignore[no-any-return]
    except Exception as ex:
        if handler is not None:
            handler(ex)
    return None


def identity(value: T) -> T:
    """
    Returns the same value.

    Args:
        value (T): The given value

    Returns:
        T: The same value
    """
    return value



from typing import Optional
from baseTest import BaseTestCase
from jstreams import Opt, equals, isTrue, strLongerThan


class TestOpt(BaseTestCase):
    def test_opt_isPresent(self) -> None:
        """
        Test opt isPresent function
        """

        val: Optional[str] = None
        self.assertFalse(Opt(val).isPresent())

        val = "test"
        self.assertTrue(Opt(val).isPresent())

        self.assertFalse(Opt(None).isPresent())

    def test_opt_get(self) -> None:
        """
        Test opt get function
        """
        self.assertThrowsExceptionOfType(lambda: Opt(None).get(), ValueError)
        self.assertIsNotNone(Opt("str").get())
        self.assertEqual(Opt("str").get(), "str")

    def test_opt_getActual(self) -> None:
        """
        Test opt getActual function
        """
        self.assertIsNotNone(Opt("str").getActual())
        self.assertEqual(Opt("str").getActual(), "str")

    def test_opt_getActual_none(self) -> None:
        """
        Test opt getActual function
        """
        self.assertIsNone(Opt("str").filter(strLongerThan(4)).getActual())

    def test_opt_getOrElse(self) -> None:
        """
        Test opt getOrElse function
        """
        self.assertIsNotNone(Opt(None).orElse("str"))
        self.assertEqual(Opt(None).orElse("str"), "str")

        self.assertIsNotNone(Opt("test").orElse("str"))
        self.assertEqual(Opt("test").orElse("str"), "test")

    def test_opt_getOrElseGet(self) -> None:
        """
        Test opt getOrElseGet function
        """
        self.assertIsNotNone(Opt(None).orElseGet(lambda: "str"))
        self.assertEqual(Opt(None).orElseGet(lambda: "str"), "str")

        self.assertIsNotNone(Opt("test").orElseGet(lambda: "str"))
        self.assertEqual(Opt("test").orElseGet(lambda: "str"), "test")

    def test_opt_stream(self) -> None:
        """
        Test opt stream function
        """
        self.assertEqual(Opt("A").stream().toList(), ["A"])
        self.assertEqual(Opt(["A"]).stream().toList(), [["A"]])

    def test_opt_flatStream(self) -> None:
        """
        Test opt flatStream function
        """
        self.assertEqual(Opt("A").flatStream().toList(), ["A"])
        self.assertEqual(Opt(["A", "B", "C"]).flatStream().toList(), ["A", "B", "C"])

    def test_opt_orElseThrow(self) -> None:
        """
        Test opt orElseThrow function
        """
        self.assertThrowsExceptionOfType(lambda: Opt(None).orElseThrow(), ValueError)
        self.assertThrowsExceptionOfType(
            lambda: Opt(None).orElseThrowFrom(lambda: Exception("Test")), Exception
        )

    def __callback_test_if_matches(self, calledStr: str) -> None:
        self.test_if_matches_result = calledStr

    def test_if_matches(self) -> None:
        """
        Test opt ifMatches function
        """
        Opt("str").ifMatches(equals("str"), self.__callback_test_if_matches)
        self.assertEqual(self.test_if_matches_result, "str")

    def test_if_matches_map(self) -> None:
        """
        Test opt ifMatchesMap function
        """
        self.assertEqual(
            Opt(True).ifMatchesMap(isTrue, lambda _: "success").get(), "success"
        )
        self.assertIsNone(
            Opt(False).ifMatchesMap(isTrue, lambda _: "success").getActual()
        )



from typing import TypeVar, Callable, Optional, Any, Generic, Protocol

from jstreams.stream import Opt

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class ErrorLog(Protocol):
    def error(self, msg: Any, *args: Any, **kwargs: Any) -> Any:
        pass


class Try(Generic[T]):
    __slots__ = (
        "__fn",
        "__thenChain",
        "__onFailure",
        "__errorLog",
        "__errorMessage",
        "__hasFailed",
        "__logger",
    )

    def __init__(self, fn: Callable[[], T]):
        self.__fn = fn
        self.__thenChain: list[Callable[[T], Any]] = []
        self.__onFailure: Optional[Callable[[BaseException], Any]] = None
        self.__errorLog: Optional[ErrorLog] = None
        self.__errorMessage: Optional[str] = None
        self.__hasFailed = False

    def withLogger(self, logger: ErrorLog) -> "Try[T]":
        self.__errorLog = logger
        return self

    def withErrorMessage(self, errorMessage: str) -> "Try[T]":
        self.__errorMessage = errorMessage
        return self

    def andThen(self, fn: Callable[[T], Any]) -> "Try[T]":
        self.__thenChain.append(fn)
        return self

    def onFailure(self, fn: Callable[[BaseException], Any]) -> "Try[T]":
        self.__onFailure = fn
        return self

    def onFailureLog(self, message: str, errorLog: ErrorLog) -> "Try[T]":
        return self.withErrorMessage(message).withLogger(errorLog)

    def get(self) -> Opt[T]:
        try:
            val = self.__fn()
            for fn in self.__thenChain:
                fn(val)
            return Opt(val)
        except Exception as e:
            self.__hasFailed = True
            if self.__onFailure is not None:
                self.__onFailure(e)
            if self.__errorLog is not None:
                if self.__errorMessage is not None:
                    self.__errorLog.error(self.__errorMessage)
                self.__errorLog.error(e, exc_info=True)
        return Opt(None)

    def hasFailed(self) -> bool:
        self.get()
        return self.__hasFailed

    @staticmethod
    def of(val: K) -> "Try[K]":
        return Try(lambda: val)



from typing import Callable, Generic, Optional, TypeVar, Union, cast, overload

from jstreams.stream import stream, Predicate
from jstreams.utils import isCallable, requireNotNull


T = TypeVar("T")
V = TypeVar("V")


class Case(Generic[T, V]):
    __slots__ = ["__matching", "__resulting"]

    def __init__(
        self,
        matching: Union[T, Callable[[T], bool], Predicate[T]],
        resulting: Union[V, Callable[[], V]],
    ) -> None:
        self.__matching = matching
        self.__resulting = resulting

    def matches(self, value: T) -> bool:
        if isinstance(self.__matching, Predicate):
            return cast(Predicate[T], self.__matching).Apply(value)
        if isCallable(self.__matching):
            return cast(Callable[[T], bool], self.__matching)(value)
        return value == self.__matching

    def result(self) -> V:
        if isCallable(self.__resulting):
            return cast(Callable[[], V], self.__resulting)()
        return cast(V, self.__resulting)


class DefaultCase(Case[T, V]):
    def __init__(
        self,
        resulting: Union[V, Callable[[], V]],
    ) -> None:
        super().__init__(lambda _: True, resulting)


class Match(Generic[T]):
    __slots__ = ["__value"]

    def __init__(self, value: T) -> None:
        self.__value = value

    @overload
    def of(self, case1: Case[T, V]) -> Optional[V]: ...

    @overload
    def of(self, case1: Case[T, V], case2: Case[T, V]) -> Optional[V]: ...

    @overload
    def of(
        self, case1: Case[T, V], case2: Case[T, V], case3: Case[T, V]
    ) -> Optional[V]: ...

    @overload
    def of(
        self, case1: Case[T, V], case2: Case[T, V], case3: Case[T, V], case4: Case[T, V]
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
        case7: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
        case7: Case[T, V],
        case8: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
        case7: Case[T, V],
        case8: Case[T, V],
        case9: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
        case7: Case[T, V],
        case8: Case[T, V],
        case9: Case[T, V],
        case10: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
        case7: Case[T, V],
        case8: Case[T, V],
        case9: Case[T, V],
        case10: Case[T, V],
        case11: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
        case7: Case[T, V],
        case8: Case[T, V],
        case9: Case[T, V],
        case10: Case[T, V],
        case11: Case[T, V],
        case12: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
        case7: Case[T, V],
        case8: Case[T, V],
        case9: Case[T, V],
        case10: Case[T, V],
        case11: Case[T, V],
        case12: Case[T, V],
        case13: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
        case7: Case[T, V],
        case8: Case[T, V],
        case9: Case[T, V],
        case10: Case[T, V],
        case11: Case[T, V],
        case12: Case[T, V],
        case13: Case[T, V],
        case14: Case[T, V],
    ) -> Optional[V]: ...

    @overload
    def of(
        self,
        case1: Case[T, V],
        case2: Case[T, V],
        case3: Case[T, V],
        case4: Case[T, V],
        case5: Case[T, V],
        case6: Case[T, V],
        case7: Case[T, V],
        case8: Case[T, V],
        case9: Case[T, V],
        case10: Case[T, V],
        case11: Case[T, V],
        case12: Case[T, V],
        case13: Case[T, V],
        case14: Case[T, V],
        case15: Case[T, V],
    ) -> Optional[V]: ...

    def of(
        self,
        case1: Case[T, V],
        case2: Optional[Case[T, V]] = None,
        case3: Optional[Case[T, V]] = None,
        case4: Optional[Case[T, V]] = None,
        case5: Optional[Case[T, V]] = None,
        case6: Optional[Case[T, V]] = None,
        case7: Optional[Case[T, V]] = None,
        case8: Optional[Case[T, V]] = None,
        case9: Optional[Case[T, V]] = None,
        case10: Optional[Case[T, V]] = None,
        case11: Optional[Case[T, V]] = None,
        case12: Optional[Case[T, V]] = None,
        case13: Optional[Case[T, V]] = None,
        case14: Optional[Case[T, V]] = None,
        case15: Optional[Case[T, V]] = None,
        case16: Optional[Case[T, V]] = None,
    ) -> Optional[V]:
        return (
            stream(
                [
                    case1,
                    case2,
                    case3,
                    case4,
                    case5,
                    case6,
                    case7,
                    case8,
                    case9,
                    case10,
                    case11,
                    case12,
                    case13,
                    case14,
                    case15,
                    case16,
                ]
            )
            .nonNull()
            .map(requireNotNull)
            .findFirst(lambda c: c.matches(self.__value))
            .map(lambda c: c.result())
            .getActual()
        )


def case(
    matching: Union[T, Callable[[T], bool], Predicate[T]],
    resulting: Union[V, Callable[[], V]],
) -> Case[T, V]:
    return Case(matching, resulting)


def match(value: T) -> Match[T]:
    return Match(value)


def matchOpt(value: Optional[T]) -> Match[Optional[T]]:
    return Match(value)


def defaultCase(resulting: Union[V, Callable[[], V]]) -> Case[T, V]:
    return DefaultCase(resulting)



from baseTest import BaseTestCase
from jstreams import (
    allNotNone,
    default,
    equals,
    isBlank,
    isIn,
    isNotIn,
    Stream,
    isNumber,
    requireNotNull,
)


class TestHelpers(BaseTestCase):
    def test_requireNotNull(self) -> None:
        """
        Test requireNotNull function
        """
        self.assertEqual(requireNotNull("str"), "str")
        self.assertThrowsExceptionOfType(lambda: requireNotNull(None), ValueError)

    def test_allSatisty(self) -> None:
        """
        Test allSatisfy function
        """
        self.assertFalse(Stream(["A", "B"]).allMatch(lambda e: e is None))
        self.assertFalse(Stream(["A", None]).allMatch(lambda e: e is None))
        self.assertTrue(Stream([None, None]).allMatch(lambda e: e is None))

    def test_areSame(self) -> None:
        """
        Test areSame function
        """
        self.assertTrue(equals([1])([1]), "Int array should be the same")
        self.assertTrue(equals(["str"])(["str"]), "String array should be the same")
        self.assertFalse(equals([1])([2]), "Int array should not be the same")
        self.assertTrue(equals({"a": "b"})({"a": "b"}), "Dict should be the same")
        self.assertTrue(
            equals({"a": "b", "c": "d"})({"a": "b", "c": "d"}),
            "Dict should be the same",
        )
        self.assertTrue(
            equals({"a": "b", "c": "d"})({"c": "d", "a": "b"}),
            "Dict should be the same",
        )
        self.assertFalse(equals({"a": "b"})({"a": "b1"}), "Dict should not be the same")

    def test_allNotNone(self) -> None:
        self.assertTrue(allNotNone(["A", "B", "C"]), "All should not be none")
        self.assertFalse(allNotNone(["A", "B", None]), "One should contain none")

    def test_isIn(self) -> None:
        self.assertTrue(isIn(["A", "B", "C"])("A"), "A should be in array")
        self.assertFalse(isIn(["A", "B", "C"])("D"), "D should not be in array")

    def test_isNotIn(self) -> None:
        self.assertFalse(isNotIn(["A", "B", "C"])("A"), "A should be in array")
        self.assertTrue(isNotIn(["A", "B", "C"])("D"), "D should not be in array")

    def test_isBlank(self) -> None:
        self.assertFalse(isBlank(["A", "B", "C"]), "Array should not be blank")
        self.assertTrue(isBlank([]), "Array should be blank")
        self.assertTrue(isBlank(None), "Object should be blank")
        self.assertTrue(isBlank(""), "Object should be blank")
        self.assertTrue(isBlank({}), "Dict should be blank")
        self.assertFalse(isBlank("Test"), "String should not be blank")
        self.assertFalse(isBlank({"a": "b"}), "Dict should not be blank")

    def test_defVal(self) -> None:
        self.assertEqual(default("str")(None), "str", "Default value should be applied")
        self.assertEqual(
            default("str")("str1"), "str1", "Given value should be applied"
        )

    def test_isNumber(self) -> None:
        self.assertTrue(isNumber(10), "10 should be a number")
        self.assertTrue(isNumber(0), "0 should be a number")
        self.assertTrue(isNumber(0.5), "0.5 should be a number")
        self.assertTrue(isNumber("10"), "10 string should be a number")
        self.assertFalse(isNumber(None), "None should not be a number")



from typing import Any, Callable, Generic, TypeVar, Union

from jstreams.stream import Predicate, predicateOf

T = TypeVar("T")
V = TypeVar("V")
K = TypeVar("K")


class Pair(Generic[T, V]):
    __slots__ = ["__left", "__right"]

    def __init__(self, left: T, right: V) -> None:
        """
        Pair constructor. The pair class is an object oriented replacement for a two value Python tuple.

        Args:
            left (T): The left value of the Pair
            right (V): The right value of the Pair
        """
        self.__left = left
        self.__right = right

    def left(self) -> T:
        return self.__left

    def right(self) -> V:
        return self.__right


class Triplet(Generic[T, V, K], Pair[T, K]):
    __slots__ = ["__middle"]

    def __init__(self, left: T, middle: V, right: K) -> None:
        """
        Triplet constructor. The triplet class is an object oriented replacement for a three value Python tuple.

        Args:
            left (T): The left value of the Triplet
            middle (V): The middle value of the Triplet
            right (K): The right value of the Triplet
        """
        super().__init__(left, right)
        self.__middle = middle

    def middle(self) -> V:
        return self.__middle


def pair(left: T, right: V) -> Pair[T, V]:
    """
    Returns a Pair object for the given values

    Args:
        left (T): The left value of the Pair
        right (V): The right value of the Pair

    Returns:
        Pair[T, V]: The Pair
    """
    return Pair(left, right)


def triplet(left: T, middle: V, right: K) -> Triplet[T, V, K]:
    """
    Returns a Triplet object for the given values

    Args:
        left (T): The left value of the Triplet
        middle (V): The middle value of the Triplet
        right (K): The right value of the Triplet

    Returns:
        Triplet[T, V, K]: The Triplet
    """
    return Triplet(left, middle, right)


def leftMatches(
    predicateArg: Union[Predicate[T], Callable[[T], bool]],
) -> Predicate[Pair[Any, Any]]:
    """
    Produces a predicate that checks if the left value of a Pair/Triplet matches the given predicate

    Args:
        predicateArg (Union[Predicate[T], Callable[[T], bool]]): The left matching predicate

    Returns:
        Predicate[Pair[T, V]]: The produced predicate
    """

    def wrap(pairArg: Pair[T, V]) -> bool:
        return predicateOf(predicateArg)(pairArg.left())

    return predicateOf(wrap)


def rightMatches(
    predicateArg: Union[Predicate[V], Callable[[V], bool]],
) -> Predicate[Pair[Any, Any]]:
    """
    Produces a predicate that checks if the right value of a Pair/Triplet matches the given predicate

    Args:
        predicateArg (Union[Predicate[V], Callable[[V], bool]]): The right matching predicate

    Returns:
        Predicate[Pair[T, V]]: The produced predicate
    """

    def wrap(pairArg: Pair[T, V]) -> bool:
        return predicateOf(predicateArg)(pairArg.right())

    return predicateOf(wrap)


def middleMatches(
    predicateArg: Union[Predicate[V], Callable[[V], bool]],
) -> Predicate[Triplet[Any, Any, Any]]:
    """
    Produces a predicate that checks if the middle value of a Triplet matches the given predicate

    Args:
        predicateArg (Union[Predicate[V], Callable[[V], bool]]): The middle matching predicate

    Returns:
        Predicate[Triplet[T, V, K]]: The produced predicate
    """

    def wrap(tripletArg: Triplet[T, V, K]) -> bool:
        return predicateOf(predicateArg)(tripletArg.middle())

    return predicateOf(wrap)



from threading import Lock
from typing import (
    Callable,
    Generic,
    Iterable,
    Optional,
    TypeVar,
    Any,
    cast,
    overload,
)
import uuid
from copy import deepcopy
from jstreams.stream import Stream
import abc

T = TypeVar("T")
A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
D = TypeVar("D")
E = TypeVar("E")
F = TypeVar("F")
G = TypeVar("G")
H = TypeVar("H")
J = TypeVar("J")
K = TypeVar("K")
L = TypeVar("L")
M = TypeVar("M")
N = TypeVar("N")
V = TypeVar("V")


ErrorHandler = Optional[Callable[[Exception], Any]]
CompletedHandler = Optional[Callable[[Optional[T]], Any]]
NextHandler = Callable[[T], Any]
DisposeHandler = Optional[Callable[[], Any]]


class RxOperator(Generic[T, V], abc.ABC):
    def __init__(self) -> None:
        pass

    def init(self) -> None:
        pass


class Pipe(Generic[T, V]):
    __slots__ = ("__operators",)

    def __init__(
        self, inputType: type[T], outputType: type[V], ops: list[RxOperator[Any, Any]]
    ) -> None:
        super().__init__()
        self.__operators: list[RxOperator[Any, Any]] = ops

    def apply(self, val: T) -> Optional[V]:
        v: Any = val
        for op in self.__operators:
            if isinstance(op, BaseFilteringOperator):
                if not op.matches(val):
                    return None
            if isinstance(op, BaseMappingOperator):
                v = op.transform(v)
        return cast(V, v)

    def clone(self) -> "Pipe[T, V]":
        return Pipe(T, V, deepcopy(self.__operators))  # type: ignore[misc]

    def init(self) -> None:
        Stream(self.__operators).each(lambda op: op.init())


class MultipleSubscriptionsException(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class ObservableSubscription(Generic[T]):
    __slots__ = (
        "__parent",
        "__onNext",
        "__onError",
        "__onCompleted",
        "__onDispose",
        "__subscriptionId",
        "__paused",
    )

    def __init__(
        self,
        parent: Any,
        onNext: NextHandler[T],
        onError: ErrorHandler = None,
        onCompleted: CompletedHandler[T] = None,
        onDispose: DisposeHandler = None,
    ) -> None:
        self.__parent = parent
        self.__onNext = onNext
        self.__onError = onError
        self.__onCompleted = onCompleted
        self.__onDispose = onDispose
        self.__subscriptionId = str(uuid.uuid4())
        self.__paused = False

    def getSubscriptionId(self) -> str:
        return self.__subscriptionId

    def onNext(self, obj: T) -> None:
        self.__onNext(obj)

    def onError(self, ex: Exception) -> None:
        if self.__onError:
            self.__onError(ex)

    def onCompleted(self, obj: Optional[T]) -> None:
        if self.__onCompleted:
            self.__onCompleted(obj)

    def isPaused(self) -> bool:
        return self.__paused

    def pause(self) -> None:
        self.__paused = True

    def resume(self) -> None:
        self.__paused = False

    def dispose(self) -> None:
        if self.__onDispose:
            self.__onDispose()

    def cancel(self) -> None:
        if hasattr(self.__parent, "cancel"):
            self.__parent.cancel(self)


class _ObservableParent(Generic[T]):
    def push(self) -> None:
        pass

    def pushToSubOnSubscribe(self, sub: ObservableSubscription[T]) -> None:
        pass


class _OnNext(Generic[T]):
    def onNext(self, val: Optional[T]) -> None:
        if not hasattr(self, "__lock"):
            self.__lock = Lock()
        with self.__lock:
            self._onNext(val)

    def _onNext(self, val: Optional[T]) -> None:
        pass


class _ObservableBase(Generic[T]):
    __slots__ = ("__subscriptions", "_parent", "_lastVal")

    def __init__(self) -> None:
        self.__subscriptions: list[ObservableSubscription[Any]] = []
        self._parent: Optional[_ObservableParent[T]] = None
        self._lastVal: Optional[T] = None

    def _notifyAllSubs(self, val: T) -> None:
        self._lastVal = val

        if self.__subscriptions is not None:
            for sub in self.__subscriptions:
                self.pushToSubscription(sub, val)

    def pushToSubscription(self, sub: ObservableSubscription[Any], val: T) -> None:
        if not sub.isPaused():
            try:
                sub.onNext(val)
            except Exception as e:
                if sub.onError is not None:
                    sub.onError(e)

    def subscribe(
        self,
        onNext: NextHandler[T],
        onError: ErrorHandler = None,
        onCompleted: CompletedHandler[T] = None,
        onDispose: DisposeHandler = None,
    ) -> ObservableSubscription[Any]:
        sub = ObservableSubscription(self, onNext, onError, onCompleted, onDispose)
        self.__subscriptions.append(sub)
        if self._parent is not None:
            self._parent.pushToSubOnSubscribe(sub)
        return sub

    def cancel(self, sub: ObservableSubscription[Any]) -> None:
        (
            Stream(self.__subscriptions)
            .filter(lambda e: e.getSubscriptionId() == sub.getSubscriptionId())
            .each(self.__subscriptions.remove)
        )

    def dispose(self) -> None:
        (Stream(self.__subscriptions).each(lambda s: s.dispose()))
        self.__subscriptions.clear()

    def pause(self, sub: ObservableSubscription[Any]) -> None:
        (
            Stream(self.__subscriptions)
            .filter(lambda e: e.getSubscriptionId() == sub.getSubscriptionId())
            .each(lambda s: s.pause())
        )

    def resume(self, sub: ObservableSubscription[Any]) -> None:
        (
            Stream(self.__subscriptions)
            .filter(lambda e: e.getSubscriptionId() == sub.getSubscriptionId())
            .each(lambda s: s.resume())
        )

    def pauseAll(self) -> None:
        (Stream(self.__subscriptions).each(lambda s: s.pause()))

    def resumePaused(self) -> None:
        (
            Stream(self.__subscriptions)
            .filter(ObservableSubscription.isPaused)
            .each(lambda s: s.resume())
        )

    def onCompleted(self, val: Optional[T]) -> None:
        (Stream(self.__subscriptions).each(lambda s: s.onCompleted(val)))
        # Clear all subscriptions. This subject is out of business
        self.dispose()

    def onError(self, ex: Exception) -> None:
        (Stream(self.__subscriptions).each(lambda s: s.onError(ex)))


class _Observable(_ObservableBase[T], _ObservableParent[T]):
    def __init__(self) -> None:
        super().__init__()


class _PipeObservable(Generic[T, V], _Observable[V]):
    __slots__ = ("__pipe", "__parent")

    def __init__(self, parent: _Observable[T], pipe: Pipe[T, V]) -> None:
        self.__pipe = pipe
        self.__parent = parent
        super().__init__()

    def subscribe(
        self,
        onNext: NextHandler[V],
        onError: ErrorHandler = None,
        onCompleted: CompletedHandler[V] = None,
        onDispose: DisposeHandler = None,
    ) -> ObservableSubscription[Any]:
        """
        Subscribe to this pipe

        Args:
            onNext (NextHandler[V]): On next handler for incoming values
            onError (ErrorHandler, optional): Error handler. Defaults to None.
            onCompleted (CompletedHandler[V], optional): Competed handler. Defaults to None.
            onDispose (DisposeHandler, optional): Dispose handler. Defaults to None.

        Returns:
            ObservableSubscription[V]: The subscription
        """
        wrappedOnNext, wrappedOnCompleted = self.__wrap(onNext, onCompleted)
        return self.__parent.subscribe(
            wrappedOnNext, onError, wrappedOnCompleted, onDispose
        )

    def __wrap(
        self, onNext: Callable[[V], Any], onCompleted: CompletedHandler[V]
    ) -> tuple[Callable[[T], Any], CompletedHandler[T]]:
        clonePipe = self.__pipe.clone()

        def onNextWrapped(val: T) -> None:
            result = clonePipe.apply(val)
            if result is not None:
                onNext(result)

        def onCompletedWrapped(val: Optional[T]) -> None:
            if val is None or onCompleted is None:
                return
            result = clonePipe.apply(val)
            if result is not None:
                onCompleted(result)

        return (onNextWrapped, onCompletedWrapped)

    def cancel(self, sub: ObservableSubscription[Any]) -> None:
        self.__parent.cancel(sub)

    def pause(self, sub: ObservableSubscription[Any]) -> None:
        self.__parent.pause(sub)

    @overload
    def pipe(
        self,
        op1: RxOperator[T, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, N],
        op10: RxOperator[N, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, N],
        op10: RxOperator[N, J],
        op11: RxOperator[J, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, N],
        op10: RxOperator[N, J],
        op11: RxOperator[J, K],
        op12: RxOperator[K, V],
    ) -> "_PipeObservable[T, V]": ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, N],
        op10: RxOperator[N, J],
        op11: RxOperator[J, K],
        op12: RxOperator[K, L],
        op13: RxOperator[L, V],
    ) -> "_PipeObservable[T, V]": ...

    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: Optional[RxOperator[A, B]] = None,
        op3: Optional[RxOperator[B, C]] = None,
        op4: Optional[RxOperator[C, D]] = None,
        op5: Optional[RxOperator[D, E]] = None,
        op6: Optional[RxOperator[E, F]] = None,
        op7: Optional[RxOperator[F, G]] = None,
        op8: Optional[RxOperator[G, H]] = None,
        op9: Optional[RxOperator[H, N]] = None,
        op10: Optional[RxOperator[N, J]] = None,
        op11: Optional[RxOperator[J, K]] = None,
        op12: Optional[RxOperator[K, L]] = None,
        op13: Optional[RxOperator[L, M]] = None,
        op14: Optional[RxOperator[M, V]] = None,
    ) -> "_PipeObservable[T, V]":
        opList = (
            Stream(
                [
                    op1,
                    op2,
                    op3,
                    op4,
                    op5,
                    op6,
                    op7,
                    op8,
                    op9,
                    op10,
                    op11,
                    op12,
                    op13,
                    op14,
                ]
            )
            .nonNull()
            .toList()
        )
        return _PipeObservable(self, Pipe(T, V, opList))  # type: ignore


class Observable(_Observable[T]):
    def __init__(self) -> None:
        super().__init__()

    @overload
    def pipe(
        self,
        op1: RxOperator[T, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, N],
        op10: RxOperator[N, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, N],
        op10: RxOperator[N, J],
        op11: RxOperator[J, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, N],
        op10: RxOperator[N, J],
        op11: RxOperator[J, K],
        op12: RxOperator[K, V],
    ) -> _PipeObservable[T, V]: ...

    @overload
    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: RxOperator[A, B],
        op3: RxOperator[B, C],
        op4: RxOperator[C, D],
        op5: RxOperator[D, E],
        op6: RxOperator[E, F],
        op7: RxOperator[F, G],
        op8: RxOperator[G, H],
        op9: RxOperator[H, N],
        op10: RxOperator[N, J],
        op11: RxOperator[J, K],
        op12: RxOperator[K, L],
        op13: RxOperator[L, V],
    ) -> _PipeObservable[T, V]: ...

    def pipe(
        self,
        op1: RxOperator[T, A],
        op2: Optional[RxOperator[A, B]] = None,
        op3: Optional[RxOperator[B, C]] = None,
        op4: Optional[RxOperator[C, D]] = None,
        op5: Optional[RxOperator[D, E]] = None,
        op6: Optional[RxOperator[E, F]] = None,
        op7: Optional[RxOperator[F, G]] = None,
        op8: Optional[RxOperator[G, H]] = None,
        op9: Optional[RxOperator[H, N]] = None,
        op10: Optional[RxOperator[N, J]] = None,
        op11: Optional[RxOperator[J, K]] = None,
        op12: Optional[RxOperator[K, L]] = None,
        op13: Optional[RxOperator[L, M]] = None,
        op14: Optional[RxOperator[M, V]] = None,
    ) -> _PipeObservable[T, V]:
        opList = (
            Stream(
                [
                    op1,
                    op2,
                    op3,
                    op4,
                    op5,
                    op6,
                    op7,
                    op8,
                    op9,
                    op10,
                    op11,
                    op12,
                    op13,
                    op14,
                ]
            )
            .nonNull()
            .toList()
        )
        return _PipeObservable(self, Pipe(T, Any, opList))  # type: ignore


class Flowable(Observable[T]):
    __slots__ = ("_values",)

    def __init__(self, values: Iterable[T]) -> None:
        super().__init__()
        self._values = values
        self._parent = self

    def push(self) -> None:
        for v in self._values:
            self._notifyAllSubs(v)

    def pushToSubOnSubscribe(self, sub: ObservableSubscription[T]) -> None:
        for v in self._values:
            self.pushToSubscription(sub, v)

    def first(self) -> Observable[T]:
        return Single(Stream(self._values).first().getActual())

    def last(self) -> Observable[T]:
        return Single(self._lastVal if self._lastVal is not None else None)


class Single(Flowable[T]):
    def __init__(self, value: Optional[T]) -> None:
        super().__init__([value] if value is not None else [])


class _SingleValueSubject(Single[T], _OnNext[T]):
    def __init__(self, value: Optional[T]) -> None:
        super().__init__(value)

    def _onNext(self, val: Optional[T]) -> None:
        if val is not None:
            self._values = [val]
            self._notifyAllSubs(val)


class BehaviorSubject(_SingleValueSubject[T]):
    def __init__(self, value: T) -> None:
        super().__init__(value)


class PublishSubject(_SingleValueSubject[T]):
    def __init__(self, typ: type[T]) -> None:
        super().__init__(None)

    def push(self) -> None:
        """
        Publish subject should not emmit anything on subscribe
        """

    def pushToSubOnSubscribe(self, sub: ObservableSubscription[T]) -> None:
        """
        Publish subject should not emmit anything on subscribe
        """


class ReplaySubject(Flowable[T], _OnNext[T]):
    __slots__ = "__valueList"

    def __init__(self, values: Iterable[T]) -> None:
        super().__init__(values)
        self.__valueList: list[T] = []

    def _onNext(self, val: Optional[T]) -> None:
        if val is not None:
            self.__valueList.append(val)
            self._notifyAllSubs(val)

    def push(self) -> None:
        super().push()
        for v in self.__valueList:
            self._notifyAllSubs(v)

    def pushToSubOnSubscribe(self, sub: ObservableSubscription[T]) -> None:
        for v in self._values:
            self.pushToSubscription(sub, v)
        for v in self.__valueList:
            self.pushToSubscription(sub, v)


class BaseFilteringOperator(RxOperator[T, T]):
    __slots__ = ("__fn",)

    def __init__(self, predicate: Callable[[T], bool]) -> None:
        self.__fn = predicate

    def matches(self, val: T) -> bool:
        return self.__fn(val)


class BaseMappingOperator(RxOperator[T, V]):
    __slots__ = ("__fn",)

    def __init__(self, mapper: Callable[[T], V]) -> None:
        self.__fn = mapper

    def transform(self, val: T) -> V:
        return self.__fn(val)


class Reduce(BaseFilteringOperator[T]):
    def __init__(self, reducer: Callable[[T, T], T]) -> None:
        """
        Reduces two consecutive values into one by applying the provided reducer function

        Args:
            reducer (Callable[[T, T], T]): Reducer function
        """
        self.__reducer = reducer
        self.__prevVal: Optional[T] = None
        super().__init__(self.__mapper)

    def init(self) -> None:
        self.__prevVal = None

    def __mapper(self, val: T) -> bool:
        if self.__prevVal is None:
            # When reducing, the first value is always returned
            self.__prevVal = val
            return True
        reduced = self.__reducer(self.__prevVal, val)
        if reduced != self.__prevVal:
            # Push and store the reduced value only if it's different than the previous value
            self.__prevVal = reduced
            return True
        return False


def rxReduce(reducer: Callable[[T, T], T]) -> RxOperator[T, T]:
    """
    Reduces two consecutive values into one by applying the provided reducer function

    Args:
        reducer (Callable[[T, T], T]): The reducer function

    Returns:
        RxOperator[T, T]: A reduce operator
    """
    return Reduce(reducer)


class Filter(BaseFilteringOperator[T]):
    def __init__(self, predicate: Callable[[T], bool]) -> None:
        """
        Allows only values that match the given predicate to flow through

        Args:
            predicate (Callable[[T], bool]): The predicate
        """
        super().__init__(predicate)


def rxFilter(predicate: Callable[[T], bool]) -> RxOperator[T, T]:
    """
    Allows only values that match the given predicate to flow through

    Args:
        predicate (Callable[[T], bool]): The predicate

    Returns:
        RxOperator[T, T]: A filter operator
    """
    return Filter(predicate)


class Map(BaseMappingOperator[T, V]):
    def __init__(self, mapper: Callable[[T], V]) -> None:
        """
        Maps a value to a differnt value/form using the mapper function

        Args:
            mapper (Callable[[T], V]): The mapper function
        """
        super().__init__(mapper)


def rxMap(mapper: Callable[[T], V]) -> RxOperator[T, V]:
    """
    Maps a value to a differnt value/form using the mapper function

    Args:
        mapper (Callable[[T], V]): The mapper function

    Returns:
        RxOperator[T, V]: A map operator
    """
    return Map(mapper)


class Take(BaseFilteringOperator[T]):
    def __init__(self, typ: type[T], count: int) -> None:
        """
        Allows only the first "count" values to flow through

        Args:
            typ (type[T]): The type of the values that will pass throgh
            count (int): The number of values that will pass through
        """
        self.__count = count
        self.__currentlyPushed = 0
        super().__init__(self.__take)

    def init(self) -> None:
        self.__currentlyPushed = 0

    def __take(self, val: T) -> bool:
        if self.__currentlyPushed >= self.__count:
            return False
        self.__currentlyPushed += 1
        return True


def rxTake(typ: type[T], count: int) -> RxOperator[T, T]:
    """
    Allows only the first "count" values to flow through

    Args:
        typ (type[T]): The type of the values that will pass throgh
        count (int): The number of values that will pass through

    Returns:
        RxOperator[T, T]: A take operator
    """
    return Take(typ, count)


class TakeWhile(BaseFilteringOperator[T]):
    def __init__(self, predicate: Callable[[T], bool]) -> None:
        """
        Allows values to pass through as long as they match the give predicate. After one value is found not matching, no other values will flow through

        Args:
            predicate (Callable[[T], bool]): The predicate
        """
        self.__fn = predicate
        self.__shouldPush = True
        super().__init__(self.__take)

    def init(self) -> None:
        self.__shouldPush = True

    def __take(self, val: T) -> bool:
        if not self.__shouldPush:
            return False
        if not self.__fn(val):
            self.__shouldPush = False
            return False
        return True


def rxTakeWhile(predicate: Callable[[T], bool]) -> RxOperator[T, T]:
    """
    Allows values to pass through as long as they match the give predicate. After one value is found not matching, no other values will flow through

    Args:
        predicate (Callable[[T], bool]): The predicate

    Returns:
        RxOperator[T, T]: A takeWhile operator
    """
    return TakeWhile(predicate)


class TakeUntil(BaseFilteringOperator[T]):
    def __init__(self, predicate: Callable[[T], bool]) -> None:
        """
        Allows values to pass through until the first value found to match the give predicate. After that, no other values will flow through

        Args:
            predicate (Callable[[T], bool]): The predicate
        """
        self.__fn = predicate
        self.__shouldPush = True
        super().__init__(self.__take)

    def init(self) -> None:
        self.__shouldPush = True

    def __take(self, val: T) -> bool:
        if not self.__shouldPush:
            return False
        if self.__fn(val):
            self.__shouldPush = False
            return False
        return True


def rxTakeUntil(predicate: Callable[[T], bool]) -> RxOperator[T, T]:
    """
    Allows values to pass through until the first value found to match the give predicate. After that, no other values will flow through

    Args:
        predicate (Callable[[T], bool]): The predicate

    Returns:
        RxOperator[T, T]: A takeUntil operator
    """
    return TakeUntil(predicate)


class Drop(BaseFilteringOperator[T]):
    def __init__(self, typ: type[T], count: int) -> None:
        """
        Blocks the first "count" values, then allows all remaining values to pass through

        Args:
            typ (type[T]): The type of the values
            count (int): The number of values to pass through
        """
        self.__count = count
        self.__currentlyDropped = 0
        super().__init__(self.__drop)

    def init(self) -> None:
        self.__currentlyDropped = 0

    def __drop(self, val: T) -> bool:
        if self.__currentlyDropped < self.__count:
            self.__currentlyDropped += 1
            return False
        return True


def rxDrop(typ: type[T], count: int) -> RxOperator[T, T]:
    """
    Blocks the first "count" values, then allows all remaining values to pass through

    Args:
        typ (type[T]): The type of the values
        count (int): The number of values to pass through

    Returns:
        RxOperator[T, T]: A drop operator
    """
    return Drop(typ, count)


class DropWhile(BaseFilteringOperator[T]):
    def __init__(self, predicate: Callable[[T], bool]) -> None:
        """
        Blocks values as long as they match the given predicate. Once a value is encountered that does not match the predicate, all remaining values will be allowed to pass through

        Args:
            predicate (Callable[[T], bool]): The predicate
        """
        self.__fn = predicate
        self.__shouldPush = False
        super().__init__(self.__drop)

    def init(self) -> None:
        self.__shouldPush = False

    def __drop(self, val: T) -> bool:
        if self.__shouldPush:
            return True

        if not self.__fn(val):
            self.__shouldPush = True
            return True
        return False


def rxDropWhile(predicate: Callable[[T], bool]) -> RxOperator[T, T]:
    """
    Blocks values as long as they match the given predicate. Once a value is encountered that does not match the predicate, all remaining values will be allowed to pass through

    Args:
        predicate (Callable[[T], bool]): The predicate

    Returns:
        RxOperator[T, T]: A dropWhile operator
    """
    return DropWhile(predicate)


class DropUntil(BaseFilteringOperator[T]):
    def __init__(self, predicate: Callable[[T], bool]) -> None:
        """
        Blocks values until the first value found that matches the given predicate. All remaining values will be allowed to pass through

        Args:
            predicate (Callable[[T], bool]): The predicate
        """
        self.__fn = predicate
        self.__shouldPush = False
        super().__init__(self.__drop)

    def init(self) -> None:
        self.__shouldPush = False

    def __drop(self, val: T) -> bool:
        if self.__shouldPush:
            return True

        if self.__fn(val):
            self.__shouldPush = True
            return True
        return False


def rxDropUntil(predicate: Callable[[T], bool]) -> RxOperator[T, T]:
    """
    Blocks values until the first value found that matches the given predicate. All remaining values will be allowed to pass through

    Args:
        predicate (Callable[[T], bool]): The given predicate

    Returns:
        RxOperator[T, T]: A dropUntil operator
    """
    return DropUntil(predicate)



from typing import Optional
from baseTest import BaseTestCase
from jstreams import (
    isNone,
    isNotNone,
    case,
    defaultCase,
    match,
    matchOpt,
    isBeweenClosed,
    isHigherThanOrEqual,
    isLessThan,
)


class TestMatch(BaseTestCase):
    def test_match_simple(self) -> None:
        val = match(5).of(
            case(1, "test1"),
            case(2, "test2"),
            case(3, "test3"),
            case(5, "test5"),
        )
        self.assertEqual(val, "test5", "Value should be correct")

    def test_match_out_of_order(self) -> None:
        val = match(5).of(
            case(1, "test1"),
            case(2, "test2"),
            case(5, "test5"),
            case(3, "test3"),
        )
        self.assertEqual(val, "test5", "Value should be correct")

    def test_match_predicate(self) -> None:
        val = match(5).of(
            case(1, "test1"),
            case(2, "test2"),
            case(isBeweenClosed(3, 6), "testBetween"),
            case(7, "test7"),
        )
        self.assertEqual(val, "testBetween", "Value should be correct")

    def test_match_predicate_supplier(self) -> None:
        val = match(5).of(
            case(1, "test1"),
            case(2, "test2"),
            case(isBeweenClosed(3, 6), lambda: "testBetween"),
            case(7, "test7"),
        )
        self.assertEqual(val, "testBetween", "Value should be correct")

    def test_match_default_case(self) -> None:
        val = match(5).of(
            case(1, "test1"),
            case(2, "test2"),
            defaultCase("default"),
        )
        self.assertEqual(val, "default", "Value should be default")

    def test_match_predicates_suppliers(self) -> None:
        val = match(8).of(
            case(isLessThan(5), lambda: str(5)),
            case(isHigherThanOrEqual(5), lambda: str(5) + "H"),
        )
        self.assertEqual(val, "5H", "Value should be correctly mapped")

    def test_match_opt_none(self) -> None:
        v: Optional[str] = None
        val = matchOpt(v).of(
            case(isNone, "None"),
            case(isNotNone, "Not none"),
        )
        self.assertEqual(val, "None", "Case should be isNone")

    def test_match_opt_not_none(self) -> None:
        v: Optional[str] = "str"
        val = matchOpt(v).of(
            case(isNone, "None"),
            case(isNotNone, "Not none"),
        )
        self.assertEqual(val, "Not none", "Case should be isNotNone")



from baseTest import BaseTestCase
from jstreams import injector
from jstreams.ioc import InjectedDependency, resolveDependencies

SUCCESS = "SUCCESS"


class TestInterface:
    def test_function(self) -> str:
        pass


class TestInterfaceImplementation(TestInterface):
    def test_function(self) -> str:
        return SUCCESS


class TestIOC(BaseTestCase):
    def setUp(self) -> None:
        injector().clear()

    def setup_interface_nq(self) -> None:
        injector().provide(TestInterface, TestInterfaceImplementation())

    def setup_interface_q(self) -> TestInterface:
        injector().provide(TestInterface, TestInterfaceImplementation(), "testName")

    def test_ioc_not_qualified(self) -> None:
        """Test dependency injection without qualifier"""
        self.assertThrowsExceptionOfType(
            lambda: injector().get(TestInterface),
            ValueError,
            "Retrieving a non existing object should throw a value error",
        )
        self.setup_interface_nq()
        self.assertIsNotNone(
            injector().find(TestInterface), "Autowired interface should not be null"
        )
        self.assertEqual(injector().get(TestInterface).test_function(), SUCCESS)

    def test_ioc_qualified(self) -> None:
        """Test dependency injection with qualifier"""
        self.assertThrowsExceptionOfType(
            lambda: injector().get(TestInterface, "testName"),
            ValueError,
            "Retrieving a non existing object should throw a value error",
        )

        self.setup_interface_q()
        self.assertIsNotNone(
            injector().find(TestInterface, "testName"),
            "Autowired interface should not be null",
        )
        self.assertEqual(
            injector().get(TestInterface, "testName").test_function(), SUCCESS
        )

    def test_autowire_public_attr(self) -> None:
        injector().provide(TestInterface, TestInterfaceImplementation())

        @resolveDependencies({"testIf": TestInterface})
        class Test:
            testIf: TestInterface

        test = Test()
        self.assertIsNotNone(test.testIf, "Attribute should have been injected")
        self.assertEqual(
            test.testIf.test_function(), SUCCESS, "Method should be properly executed"
        )

    def test_autowire_protected_attr(self) -> None:
        injector().provide(TestInterface, TestInterfaceImplementation())

        @resolveDependencies({"_testIf": TestInterface})
        class Test:
            _testIf: TestInterface

            def getTestIf(self) -> TestInterface:
                return self._testIf

        test = Test()
        self.assertIsNotNone(test.getTestIf(), "Attribute should have been injected")
        self.assertEqual(
            test.getTestIf().test_function(),
            SUCCESS,
            "Method should be properly executed",
        )

    def test_injected_dependency_class(self) -> None:
        injector().provide(TestInterface, TestInterfaceImplementation())

        class Test:
            def __init__(self):
                self.dep = InjectedDependency(TestInterface)

        test = Test()
        self.assertIsNotNone(test.dep)
        self.assertEqual(test.dep.get().test_function(), SUCCESS)

    def test_injected_dependency_class_fail(self) -> None:
        class Test:
            def __init__(self):
                self.dep = InjectedDependency(TestInterface)

        test = Test()
        self.assertIsNotNone(test.dep)
        self.assertThrowsExceptionOfType(
            test.dep.get,
            ValueError,
            "Should throw error when dependency is forced and not present",
        )



from baseTest import BaseTestCase
from jstreams import isNone, isNotNone, not_, predicateOf


class TestNot(BaseTestCase):
    def test_not_fn(self) -> None:
        self.assertFalse(
            not_(isNone)(None), "Not isNone applied to None should be False"
        )
        self.assertTrue(
            not_(isNotNone)(None), "Not isNotNone applied to None should be True"
        )

    def test_not_predicate(self) -> None:
        self.assertFalse(
            not_(predicateOf(isNone))(None),
            "Not isNone applied to None should be False",
        )
        self.assertTrue(
            not_(predicateOf(isNotNone))(None),
            "Not isNotNone applied to None should be True",
        )



from typing import Any
from baseTest import BaseTestCase
from jstreams import Try


class CallRegister:
    def __init__(self):
        self.mth1Called = False
        self.mth2Called = False
        self.mth3Called = False
        self.mth4Called = False
        self.errorLogged = False

    def mth1(self, e: Any) -> None:
        self.mth1Called = True

    def mth2(self, e: Any) -> None:
        self.mth2Called = True

    def mth3(self, e: Any) -> None:
        self.mth3Called = True

    def mth4(self, e: Any) -> None:
        self.mth4Called = True

    def error(self, msg, *args, **kwargs):
        self.errorLogged = True


class TestTry(BaseTestCase):
    def noThrow(self) -> str:
        return "str"

    def throw(self) -> str:
        raise ValueError("Test")

    def processThrow(self, e: str) -> None:
        raise ValueError("Test")

    def test_try(self) -> None:
        mock = CallRegister()
        self.assertEqual(
            Try(self.noThrow).andThen(mock.mth1).andThen(mock.mth2).get().get(),
            "str",
        )
        self.assertTrue(mock.mth1Called)
        self.assertTrue(mock.mth2Called)

    def test_try_with_error_on_initial(self) -> None:
        mock = CallRegister()
        self.assertIsNone(Try(self.throw).andThen(mock.mth1).get().getActual())
        self.assertFalse(mock.mth1Called)

    def test_try_with_error_on_chain(self) -> None:
        self.assertIsNone(
            Try(self.noThrow).andThen(self.processThrow).get().getActual()
        )

    def test_try_with_error_on_init_and_onFailure(self) -> None:
        mock = CallRegister()
        self.assertIsNone(
            Try(self.throw).andThen(mock.mth1).onFailure(mock.mth2).get().getActual()
        )
        self.assertFalse(mock.mth1Called)
        self.assertTrue(mock.mth2Called)

    def test_try_with_error_on_chain_and_onFailure(self) -> None:
        mock = CallRegister()
        self.assertIsNone(
            Try(self.noThrow)
            .andThen(self.processThrow)
            .onFailure(mock.mth1)
            .get()
            .getActual()
        )
        self.assertTrue(mock.mth1Called)

    def test_try_with_error_on_chain_and_onFailureLog(self) -> None:
        mock = CallRegister()
        self.assertIsNone(
            Try(self.noThrow)
            .andThen(self.processThrow)
            .onFailureLog("Test", mock)
            .get()
            .getActual()
        )
        self.assertTrue(mock.errorLogged)



from typing import (
    Callable,
    Iterable,
    Any,
    Iterator,
    Optional,
    Sized,
    TypeVar,
    Generic,
    cast,
    Union,
)
from abc import ABC, abstractmethod


T = TypeVar("T")
V = TypeVar("V")
K = TypeVar("K")
C = TypeVar("C")


class Predicate(ABC, Generic[T]):
    @abstractmethod
    def Apply(self, value: T) -> bool:
        """
        Apply a condition to a given value.

        Args:
            value (T): The value

        Returns:
            bool: True if the value matches, False otherwise
        """

    def Or(self, other: Union[Callable[[T], bool], "Predicate[T]"]) -> "Predicate[T]":
        return predicateOf(lambda v: self.Apply(v) or predicateOf(other).Apply(v))

    def And(self, other: Union[Callable[[T], bool], "Predicate[T]"]) -> "Predicate[T]":
        return predicateOf(lambda v: self.Apply(v) and predicateOf(other).Apply(v))

    def __call__(self, value: T) -> bool:
        return self.Apply(value)


class PredicateWith(ABC, Generic[T, K]):
    @abstractmethod
    def Apply(self, value: T, withValue: K) -> bool:
        """
        Apply a condition to two given values.

        Args:
            value (T): The value
            withValue (K): The second value

        Returns:
            bool: True if the values matche the predicate, False otherwise
        """

    def Or(self, other: "PredicateWith[T, K]") -> "PredicateWith[T, K]":
        return predicateWithOf(lambda v, k: self.Apply(v, k) or other.Apply(v, k))

    def And(self, other: "PredicateWith[T, K]") -> "PredicateWith[T, K]":
        return predicateWithOf(lambda v, k: self.Apply(v, k) and other.Apply(v, k))

    def __call__(self, value: T, withValue: K) -> bool:
        return self.Apply(value, withValue)


class _WrapPredicate(Predicate[T]):
    __slots__ = ["__predicateFn"]

    def __init__(self, fn: Callable[[T], bool]) -> None:
        self.__predicateFn = fn

    def Apply(self, value: T) -> bool:
        return self.__predicateFn(value)


class _WrapPredicateWith(PredicateWith[T, K]):
    __slots__ = ["__predicateFn"]

    def __init__(self, fn: Callable[[T, K], bool]) -> None:
        self.__predicateFn = fn

    def Apply(self, value: T, withValue: K) -> bool:
        return self.__predicateFn(value, withValue)


class Mapper(ABC, Generic[T, V]):
    @abstractmethod
    def Map(self, value: T) -> V:
        """
        Maps the given value, to a new value of maybe a different type.

        Args:
            value (T): The given value

        Returns:
            V: The produced value
        """

    def __call__(self, value: T) -> V:
        return self.Map(value)


class MapperWith(ABC, Generic[T, K, V]):
    @abstractmethod
    def Map(self, value: T, withValue: K) -> V:
        """
        Maps the given two values, to a new value.

        Args:
            value (T): The given value
            withValue (K): The scond value

        Returns:
            V: The produced value
        """

    def __call__(self, value: T, withValue: K) -> V:
        return self.Map(value, withValue)


class _WrapMapper(Mapper[T, V]):
    __slots__ = ["__mapper"]

    def __init__(self, mapper: Callable[[T], V]) -> None:
        self.__mapper = mapper

    def Map(self, value: T) -> V:
        return self.__mapper(value)


class _WrapMapperWith(MapperWith[T, K, V]):
    __slots__ = ["__mapper"]

    def __init__(self, mapper: Callable[[T, K], V]) -> None:
        self.__mapper = mapper

    def Map(self, value: T, withValue: K) -> V:
        return self.__mapper(value, withValue)


class Reducer(ABC, Generic[T]):
    @abstractmethod
    def Reduce(self, a: T, b: T) -> T:
        """
        Reduce two values to a single one.

        Args:
            a (T): The first value
            b (T): The second value

        Returns:
            T: The reduced value
        """

    def __call__(self, a: T, b: T) -> T:
        return self.Reduce(a, b)


class _WrapReducer(Reducer[T]):
    __slots__ = ["__reducer"]

    def __init__(self, reducer: Callable[[T, T], T]) -> None:
        self.__reducer = reducer

    def Reduce(self, a: T, b: T) -> T:
        return self.__reducer(a, b)


def reducerOf(reducer: Union[Reducer[T], Callable[[T, T], T]]) -> Reducer[T]:
    if isinstance(reducer, Reducer):
        return reducer
    return _WrapReducer(reducer)


def mapperOf(mapper: Union[Mapper[T, V], Callable[[T], V]]) -> Mapper[T, V]:
    """
    If the value passed is a mapper, it is returned without changes.
    If a function is passed, it will be wrapped into a Mapper object.

    Args:
        mapper (Union[Mapper[T, V], Callable[[T], V]]): The mapper

    Returns:
        Mapper[T, V]: The produced mapper
    """
    if isinstance(mapper, Mapper):
        return mapper
    return _WrapMapper(mapper)


def mapperWithOf(
    mapper: Union[MapperWith[T, K, V], Callable[[T, K], V]],
) -> MapperWith[T, K, V]:
    """
    If the value passed is a mapper, it is returned without changes.
    If a function is passed, it will be wrapped into a Mapper object.


    Args:
        mapper (Union[MapperWith[T, K, V], Callable[[T, K], V]]): The mapper

    Returns:
        MapperWith[T, K, V]: The produced mapper
    """
    if isinstance(mapper, MapperWith):
        return mapper
    return _WrapMapperWith(mapper)


def predicateOf(predicate: Union[Predicate[T], Callable[[T], bool]]) -> Predicate[T]:
    """
    If the value passed is a predicate, it is returned without any changes.
    If a function is passed, it will be wrapped into a Predicate object.

    Args:
        predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

    Returns:
        Predicate[T]: The produced predicate
    """
    if isinstance(predicate, Predicate):
        return predicate
    return _WrapPredicate(predicate)


def predicateWithOf(
    predicate: Union[PredicateWith[T, K], Callable[[T, K], bool]],
) -> PredicateWith[T, K]:
    """
    If the value passed is a predicate, it is returned without any changes.
    If a function is passed, it will be wrapped into a Predicate object.

    Args:
        predicate (Union[PredicateWith[T, K], Callable[[T, K], bool]]): The predicate

    Returns:
        PredicateWith[T, K]: The produced predicate
    """
    if isinstance(predicate, PredicateWith):
        return predicate
    return _WrapPredicateWith(predicate)


def isEmptyOrNone(
    obj: Union[list[Any], dict[Any, Any], str, None, Any, Iterable[Any]],
) -> bool:
    """
    Checkes whether the given object is either None, or is empty.
    For str and Sized objects, besides the None check, the len(obj) == 0 is also applied

    Args:
        obj (Union[list[Any], dict[Any, Any], str, None, Any, Iterable[Any]]): The object

    Returns:
        bool: True if empty or None, False otherwise
    """
    if obj is None:
        return True

    if isinstance(obj, Iterable):
        for _ in obj:
            return False
        return True

    if isinstance(obj, Sized):
        return len(obj) == 0

    return False


def cmpToKey(mycmp: Callable[[C, C], int]) -> type:
    """Convert a cmp= function into a key= function"""

    class Key(Generic[C]):  # type: ignore[misc]
        __slots__ = ["obj"]

        def __init__(self, obj: C) -> None:
            self.obj = obj

        def __lt__(self, other: "Key") -> bool:
            return mycmp(self.obj, other.obj) < 0

        def __gt__(self, other: "Key") -> bool:
            return mycmp(self.obj, other.obj) > 0

        def __eq__(self, other: object) -> bool:
            if not isinstance(other, Key):
                return NotImplemented
            return mycmp(self.obj, other.obj) == 0

        def __le__(self, other: "Key") -> bool:
            return mycmp(self.obj, other.obj) <= 0

        def __ge__(self, other: "Key") -> bool:
            return mycmp(self.obj, other.obj) >= 0

    return Key


def each(target: Optional[Iterable[T]], action: Callable[[T], Any]) -> None:
    """
    Executes an action on each element of the given iterable

    Args:
        target (Optional[Iterable[T]]): The target iterable
        action (Callable[[T], Any]): The action to be executed
    """
    if target is None:
        return

    for el in target:
        action(el)


def findFirst(
    target: Optional[Iterable[T]], predicate: Union[Predicate[T], Callable[[T], bool]]
) -> Optional[T]:
    """
    Retrieves the first element of the given iterable that matches the given predicate

    Args:
        target (Optional[Iterable[T]]): The target iterable
        predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

    Returns:
        Optional[T]: The first matching element, or None if no element matches the predicate
    """
    if target is None:
        return None

    for el in target:
        if predicateOf(predicate).Apply(el):
            return el
    return None


def mapIt(
    target: Iterable[T], mapper: Union[Mapper[T, V], Callable[[T], V]]
) -> list[V]:
    """
    Maps each element of an iterable to a new object produced by the given mapper

    Args:
        target (Iterable[T]): The target iterable
        mapper (Union[Mapper[T, V], Callable[[T], V]]): The mapper

    Returns:
        list[V]: The mapped elements
    """
    if target is None:
        return []
    mapperObj = mapperOf(mapper)
    return [mapperObj.Map(el) for el in target]


def flatMap(
    target: Iterable[T],
    mapper: Union[Mapper[T, Iterable[V]], Callable[[T], Iterable[V]]],
) -> list[V]:
    """
    Returns a flattened map. The mapper function is called for each element of the target
    iterable, then all elements are added to a result list.
    Ex: flatMap([1, 2], lambda x: [x, x + 1]) returns [1, 2, 2, 3]

    Args:
        target (Iterable[T]): The target iterable
        mapper (Union[Mapper[T, V], Callable[[T], V]]): The mapper

    Returns:
        list[V]: The resulting flattened map
    """
    ret: list[V] = []
    if target is None:
        return ret

    mapperObj = mapperOf(mapper)

    for el in target:
        mapped = mapperObj.Map(el)
        each(mapped, ret.append)
    return ret


def matching(
    target: Iterable[T], predicate: Union[Predicate[T], Callable[[T], bool]]
) -> list[T]:
    """
    Returns all elements of the target iterable that match the given predicate

    Args:
        target (Iterable[T]): The target iterable
        predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

    Returns:
        list[T]: The matching elements
    """
    ret: list[T] = []
    if target is None:
        return ret

    for el in target:
        if predicateOf(predicate).Apply(el):
            ret.append(el)
    return ret


def takeWhile(
    target: Iterable[T], predicate: Union[Predicate[T], Callable[[T], bool]]
) -> list[T]:
    """
    Returns the first batch of elements matching the predicate. Once an element
    that does not match the predicate is found, the function will return

    Args:
        target (Iterable[T]): The target iterable
        predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

    Returns:
        list[T]: The result list
    """
    ret: list[T] = []
    if target is None:
        return ret

    for el in target:
        if predicateOf(predicate).Apply(el):
            ret.append(el)
        else:
            break
    return ret


def dropWhile(
    target: Iterable[T], predicate: Union[Predicate[T], Callable[[T], bool]]
) -> list[T]:
    """
    Returns the target iterable elements without the first elements that match the
    predicate. Once an element that does not match the predicate is found,
    the function will start adding the remaining elements to the result list

    Args:
        target (Iterable[T]): The target iterable
        predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

    Returns:
        list[T]: The result list
    """
    ret: list[T] = []
    if target is None:
        return ret

    index = 0

    for el in target:
        if predicateOf(predicate).Apply(el):
            index += 1
        else:
            break
    return list(target)[index:]


def reduce(
    target: Iterable[T], reducer: Union[Reducer[T], Callable[[T, T], T]]
) -> Optional[T]:
    """
    Reduces an iterable to a single value. The reducer function takes two values and
    returns only one. This function can be used to find min or max from a stream of ints.

    Args:
        reducer (Union[Reducer[T], Callable[[T, T], T]]): The reducer

    Returns:
        Optional[T]: The resulting optional
    """

    if target is None:
        return None

    elemList = list(target)
    if len(elemList) == 0:
        return None

    result: T = elemList[0]
    reducerObj = reducerOf(reducer)
    for el in elemList:
        result = reducerObj.Reduce(el, result)
    return result


def isNotNone(element: Optional[T]) -> bool:
    """
    Checks if the given element is not None. This function is meant to be used
    instead of lambdas for non null checks

    Args:
        element (Optional[T]): The given element

    Returns:
        bool: True if element is not None, False otherwise
    """
    return element is not None


def dictUpdate(target: dict[K, V], key: K, value: V) -> None:
    target[key] = value


def sort(target: list[T], comparator: Callable[[T, T], int]) -> list[T]:
    """
    Returns a list with the elements sorted according to the comparator function.
    CAUTION: This method will actually iterate the entire iterable, so if you're using
    infinite generators, calling this method will block the execution of the program.

    Args:
        comparator (Callable[[T, T], int]): The comparator function

    Returns:
        list[T]: The resulting list
    """

    return sorted(target, key=cmpToKey(comparator))


class Opt(Generic[T]):
    __slots__ = ("__val",)
    __NONE: "Optional[Opt[Any]]" = None

    def __init__(self, val: Optional[T]) -> None:
        self.__val = val

    def __getNone(self) -> "Opt[T]":
        if Opt.__NONE is None:
            Opt.__NONE = Opt(None)
        return cast(Opt[T], Opt.__NONE)

    def get(self) -> T:
        """
        Returns the value of the Opt object if present, otherwise will raise a ValueError

        Raises:
            ValueError: Error raised when the value is None

        Returns:
            T: The value
        """
        if self.__val is None:
            raise ValueError("Object is None")
        return self.__val

    def getActual(self) -> Optional[T]:
        """
        Returns the actual value of the Opt without raising any errors

        Returns:
            Optional[T]: The value
        """
        return self.__val

    def orElse(self, val: T) -> T:
        """
        Returns the value of the Opt if present, otherwise return the given parameter as a fallback.
        This functiona should be used when the given fallback is a constant or it does not require
        heavy computation

        Args:
            val (T): The fallback value

        Returns:
            T: The return value
        """
        return self.__val if self.__val is not None else val

    def orElseOpt(self, val: Optional[T]) -> Optional[T]:
        """
        Returns the value of the Opt if present, otherwise return the given parameter as a fallback.
        This functiona should be used when the given fallback is a constant or it does not require
        heavy computation

        Args:
            val (Optional[T]): The optional fallback value

        Returns:
            T: The return value
        """
        return self.__val if self.__val is not None else val

    def orElseGetOpt(self, supplier: Callable[[], Optional[T]]) -> Optional[T]:
        """
        Returns the value of the Opt if present, otherwise it will call the supplier
        function and return that value. This function is useful when the fallback value
        is compute heavy and should only be called when the value of the Opt is None

        Args:
            supplier (Callable[[], T]): The mandatory return supplier

        Returns:
            Optional[T]: The resulting value
        """
        return self.__val if self.__val is not None else supplier()

    def orElseGet(self, supplier: Callable[[], T]) -> T:
        """
        Returns the value of the Opt if present, otherwise it will call the supplier
        function and return that value. This function is useful when the fallback value
        is compute heavy and should only be called when the value of the Opt is None

        Args:
            supplier (Callable[[], T]): The mandatory value supplier

        Returns:
            Optional[T]: _description_
        """
        return self.__val if self.__val is not None else supplier()

    def isPresent(self) -> bool:
        """
        Returns whether the Opt is present

        Returns:
            bool: True if the Opt has a non null value, False otherwise
        """
        return self.__val is not None

    def isEmpty(self) -> bool:
        """
        Returns whether the Opt is empty

        Returns:
            bool: True if the Opt value is None, False otherwise
        """
        return self.__val is None

    def ifPresent(self, action: Callable[[T], Any]) -> "Opt[T]":
        """
        Executes an action on the value of the Opt if the value is present

        Args:
            action (Callable[[T], Any]): The action
        Returns:
            Opt[T]: This optional
        """
        if self.__val is not None:
            action(self.__val)
        return self

    def ifPresentWith(self, withVal: K, action: Callable[[T, K], Any]) -> "Opt[T]":
        """
        Executes an action on the value of the Opt if the value is present, by providing
        the action an additional parameter

        Args:
            withVal (K): The additional parameter
            action (Callable[[T, K], Any]): The action
        Returns:
            Opt[T]: This optional
        """
        if self.__val is not None:
            action(self.__val, withVal)
        return self

    def ifNotPresent(self, action: Callable[[], Any]) -> "Opt[T]":
        """
        Executes an action on if the value is not present

        Args:
            action (Callable[[], Any]): The action
        Returns:
            Opt[T]: This optional
        """
        if self.__val is None:
            action()
        return self

    def ifNotPresentWith(self, withVal: K, action: Callable[[K], Any]) -> "Opt[T]":
        """
        Executes an action on if the value is not present, by providing
        the action an additional parameter

        Args:
            withVal (K): The additional parameter
            action (Callable[[K], Any]): The action
        Returns:
            Opt[T]: This optional
        """
        if self.__val is None:
            action(withVal)
        return self

    def ifPresentOrElse(
        self, action: Callable[[T], Any], emptyAction: Callable[[], Any]
    ) -> "Opt[T]":
        """
        Executes an action on the value of the Opt if the value is present, or executes
        the emptyAction if the Opt is empty

        Args:
            action (Callable[[T], Any]): The action to be executed when present
            emptyAction (Callable[[], Any]): The action to be executed when empty
        Returns:
            Opt[T]: This optional
        """
        if self.__val is not None:
            action(self.__val)
        else:
            emptyAction()
        return self

    def ifPresentOrElseWith(
        self, withVal: K, action: Callable[[T, K], Any], emptyAction: Callable[[K], Any]
    ) -> "Opt[T]":
        """
        Executes an action on the value of the Opt by providing the actions an additional parameter,
        if the value is present, or executes the emptyAction if the Opt is empty

        Args:
            withVal (K): The additional parameter
            action (Callable[[T, K], Any]): The action to be executed when present
            emptyAction (Callable[[K], Any]): The action to be executed when empty
        """
        if self.__val is not None:
            action(self.__val, withVal)
        else:
            emptyAction(withVal)
        return self

    def filter(self, predicate: Union[Predicate[T], Callable[[T], bool]]) -> "Opt[T]":
        """
        Returns the filtered value of the Opt if it matches the given predicate

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

        Returns:
            Opt[T]: The resulting Opt
        """
        if self.__val is None:
            return self
        if predicateOf(predicate).Apply(self.__val):
            return self
        return self.__getNone()

    def filterWith(
        self, withVal: K, predicate: Union[PredicateWith[T, K], Callable[[T, K], bool]]
    ) -> "Opt[T]":
        """
        Returns the filtered value of the Opt if it matches the given predicate, by
        providing the predicat with an additional value

        Args:
            withVal (K): the additional value
            predicate (Union[PredicateWith[T, K], Callable[[T, K], bool]]): The predicate

        Returns:
            Opt[T]: The resulting Opt
        """
        if self.__val is None:
            return self
        if predicateWithOf(predicate).Apply(self.__val, withVal):
            return self
        return self.__getNone()

    def map(self, mapper: Union[Mapper[T, V], Callable[[T], V]]) -> "Opt[V]":
        """
        Maps the Opt value into another Opt by applying the mapper function

        Args:
            mapper (Callable[[T], V]): The mapper function

        Returns:
            Opt[V]: The resulting Opt
        """
        if self.__val is None:
            return cast(Opt[V], self.__getNone())
        return Opt(mapperOf(mapper).Map(self.__val))

    def mapWith(
        self, withVal: K, mapper: Union[MapperWith[T, K, V], Callable[[T, K], V]]
    ) -> "Opt[V]":
        """
        Maps the Opt value into another Opt by applying the mapper function with an additional parameter

        Args:
            withVal (K): The additional parameter
            mapper (Callable[[T, K], V]): The mapper function

        Returns:
            Opt[V]: The resulting Opt
        """
        if self.__val is None:
            return cast(Opt[V], self.__getNone())
        return Opt(mapperWithOf(mapper).Map(self.__val, withVal))

    def orElseGetWith(self, withVal: K, supplier: Callable[[K], T]) -> "Opt[T]":
        """
        Returns this Opt if present, otherwise will return the supplier result with
        the additional parameter

        Args:
            withVal (K): The additional parameter
            supplier (Callable[[K], T]): The supplier

        Returns:
            Opt[T]: The resulting Opt
        """
        return self.orElseGetWithOpt(withVal, supplier)

    def orElseGetWithOpt(
        self, withVal: K, supplier: Callable[[K], Optional[T]]
    ) -> "Opt[T]":
        """
        Returns this Opt if present, otherwise will return the supplier result with
        the additional parameter

        Args:
            withVal (K): The additional parameter
            supplier (Callable[[K], Optional[T]]): The supplier

        Returns:
            Opt[T]: The resulting Opt
        """
        if self.isPresent():
            return self
        return Opt(supplier(withVal))

    def ifMatches(
        self,
        predicate: Union[Predicate[T], Callable[[T], bool]],
        action: Callable[[T], Any],
    ) -> "Opt[T]":
        """
        Executes the given action on the value of this Opt, if the value is present and
        matches the given predicate. Returns the same Opt

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate
            action (Callable[[T], Any]): The action to be executed

        Returns:
            Opt[T]: The same Opt
        """
        if self.__val is not None and predicateOf(predicate).Apply(self.__val):
            action(self.__val)
        return self

    def ifMatchesOpt(
        self,
        predicate: Union[Predicate[Optional[T]], Callable[[Optional[T]], bool]],
        action: Callable[[Optional[T]], Any],
    ) -> "Opt[T]":
        """
        Executes the given action on the value of this Opt, regardless of whether the value
        is present, if the value matches the given predicate. Returns the same Opt

        Args:
            predicate (Union[Predicate[Optional[T]], Callable[[Optional[T]], bool]]): The predicate
            action (Callable[[Optional[T]], Any]): The action to be executed

        Returns:
            Opt[T]: The same Opt
        """
        if predicateOf(predicate).Apply(self.__val):
            action(self.__val)
        return self

    def stream(self) -> "Stream[T]":
        """
        Returns a Stream containing the current Opt value

        Returns:
            Stream[T]: The resulting Stream
        """
        if self.__val is not None:
            return Stream([self.__val])
        return Stream([])

    def flatStream(self) -> "Stream[T]":
        """
        Returns a Stream containing the current Opt value if the value
        is not an Iterable, or a Stream containing all the values in
        the Opt if the Opt contains an iterable

        Returns:
            Stream[T]: The resulting Stream
        """
        if self.__val is not None:
            if isinstance(self.__val, Iterable):
                return Stream(self.__val)
            return Stream([self.__val])
        return Stream([])

    def orElseThrow(self) -> T:
        """
        Returns the value of the Opt or raise a value error

        Raises:
            ValueError: The value error

        Returns:
            T: The value
        """
        if self.__val is not None:
            return self.__val
        raise ValueError("Object is None")

    def orElseThrowFrom(self, exceptionSupplier: Callable[[], BaseException]) -> T:
        """
        Returns the value of the Opt or raise an exeption provided by the exception supplier

        Args:
            exceptionSupplier (Callable[[], BaseException]): The exception supplier

        Raises:
            exception: The generated exception

        Returns:
            T: The value
        """
        if self.__val is not None:
            return self.__val
        raise exceptionSupplier()

    def ifPresentMap(
        self,
        isPresentMapper: Union[Mapper[T, V], Callable[[T], V]],
        orElseSupplier: Callable[[], Optional[V]],
    ) -> "Opt[V]":
        """
        If the optional value is present, returns the value mapped by isPresentMapper wrapped in an Opt.
        If the optional value is not present, returns the value produced by orElseSupplier

        Args:
            isPresentMapper (Union[Mapper[T, V], Callable[[T], V]]): The presence mapper
            orElseSupplier (Callable[[], Optional[V]]): The missing value producer

        Returns:
            Opt[V]: An optional
        """
        if self.__val is None:
            return Opt(orElseSupplier())
        return Opt(mapperOf(isPresentMapper).Map(self.__val))

    def ifPresentMapWith(
        self,
        withVal: K,
        isPresentMapper: Union[MapperWith[T, K, V], Callable[[T, K], V]],
        orElseSupplier: Callable[[K], Optional[V]],
    ) -> "Opt[V]":
        """
        If the optional value is present, returns the value mapped by isPresentMapper wrapped in an Opt.
        If the optional value is not present, returns the value produced by orElseSupplier.
        In addition to ifPresentMap, this method also passes the withVal param to the mapper and supplier

        Args:
            withVal (K): The additional mapper value
            isPresentMapper (Union[MapperWith[T, K, V],  Callable[[T, K], V]]): The presence mapper
            orElseSupplier (Callable[[K], V]): The missing value producer

        Returns:
            Opt[V]: An optional
        """
        if self.__val is None:
            return Opt(orElseSupplier(withVal))
        return Opt(mapperWithOf(isPresentMapper).Map(self.__val, withVal))

    def instanceOf(self, classType: type) -> "Opt[T]":
        """
        Equivalent of Opt.filter(lambda val: isinstance(val, classType))

        Args:
            classType (type): The class type

        Returns:
            Opt[T]: An optional
        """
        if isinstance(self.__val, classType):
            return self
        return self.__getNone()

    def cast(self, classType: type[V]) -> "Opt[V]":
        """
        Equivalent of Opt.map(lambda val: cast(classType, val))

        Args:
            classType (type[V]): The class type of the new optional

        Returns:
            Opt[V]: An optional
        """
        return Opt(cast(V, self.__val))

    def ifMatchesMap(
        self,
        predicate: Union[Predicate[T], Callable[[T], bool]],
        mapper: Union[Mapper[T, Optional[V]], Callable[[T], Optional[V]]],
    ) -> "Opt[V]":
        """
        If the optional value is present and matches the given predicate, returns the value mapped
        by mapper wrapped in an Opt.
        If the optional value is not present, returns an empty Opt.

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate
            mapper (Union[Mapper[T, V], Callable[[T], Optional[V]]]): The the mapper

        Returns:
            Opt[V]: An optional
        """
        if self.__val is not None and predicateOf(predicate).Apply(self.__val):
            return Opt(mapperOf(mapper).Map(self.__val))
        return cast(Opt[V], self.__getNone())

    def ifMatchesMapWith(
        self,
        withVal: K,
        predicate: Union[PredicateWith[T, K], Callable[[T, K], bool]],
        mapper: Union[MapperWith[T, K, Optional[V]], Callable[[T, K], Optional[V]]],
    ) -> "Opt[V]":
        """
        If the optional value is present and matches the given predicate, returns the value mapped by mapper wrapped in an Opt.
        If the optional value is not present, returns an empty Opt.
        In addition to ifMatchesMap, this method also passes the withVal param to the mapper and supplier

        Args:
            withVal (K): The additional mapper value
            predicate (Union[PredicateWith[T, K], Callable[[T, K], bool]]): The predicate
            mapper (Union[MapperWith[T, K, Optional[V]], Callable[[T, K], Optional[V]]]): The mapper

        Returns:
            Opt[V]: An optional
        """
        if self.__val is not None and predicateWithOf(predicate).Apply(
            self.__val, withVal
        ):
            return Opt(mapperWithOf(mapper).Map(self.__val, withVal))
        return cast(Opt[V], self.__getNone())


class ClassOps:
    __slots__ = ("__classType",)

    def __init__(self, classType: type) -> None:
        self.__classType = classType

    def instanceOf(self, obj: Any) -> bool:
        return isinstance(obj, self.__classType)

    def subClassOf(self, typ: type) -> bool:
        return issubclass(typ, self.__classType)


class _GenericIterable(ABC, Generic[T], Iterator[T], Iterable[T]):
    __slots__ = ("_iterable", "_iterator")

    def __init__(self, it: Iterable[T]) -> None:
        self._iterable = it
        self._iterator = self._iterable.__iter__()

    def _prepare(self) -> None:
        pass

    def __iter__(self) -> Iterator[T]:
        self._iterator = self._iterable.__iter__()
        self._prepare()
        return self


class _FilterIterable(_GenericIterable[T]):
    __slots__ = ("__predicate",)

    def __init__(self, it: Iterable[T], predicate: Predicate[T]) -> None:
        super().__init__(it)
        self.__predicate = predicate

    def __next__(self) -> T:
        while True:
            nextObj = self._iterator.__next__()
            if self.__predicate.Apply(nextObj):
                return nextObj


class _CastIterable(Generic[T, V], Iterator[T], Iterable[T]):
    __slots__ = ("__iterable", "__iterator", "__tp")

    def __init__(self, it: Iterable[V], typ: type[T]) -> None:
        self.__iterable = it
        self.__iterator = self.__iterable.__iter__()
        self.__tp = typ

    def __iter__(self) -> Iterator[T]:
        self.__iterator = self.__iterable.__iter__()
        return self

    def __next__(self) -> T:
        nextObj = self.__iterator.__next__()
        return cast(T, nextObj)


class _SkipIterable(_GenericIterable[T]):
    __slots__ = ("__count",)

    def __init__(self, it: Iterable[T], count: int) -> None:
        super().__init__(it)
        self.__count = count

    def _prepare(self) -> None:
        try:
            count = 0
            while count < self.__count:
                self._iterator.__next__()
                count += 1
        except StopIteration:
            pass

    def __next__(self) -> T:
        return self._iterator.__next__()


class _LimitIterable(_GenericIterable[T]):
    __slots__ = ("__count", "__currentCount")

    def __init__(self, it: Iterable[T], count: int) -> None:
        super().__init__(it)
        self.__count = count
        self.__currentCount = 0

    def _prepare(self) -> None:
        self.__currentCount = 0

    def __next__(self) -> T:
        if self.__currentCount >= self.__count:
            raise StopIteration()

        obj = self._iterator.__next__()
        self.__currentCount += 1
        return obj


class _TakeWhileIterable(_GenericIterable[T]):
    __slots__ = ("__predicate", "__done")

    def __init__(self, it: Iterable[T], predicate: Predicate[T]) -> None:
        super().__init__(it)
        self.__done = False
        self.__predicate = predicate

    def _prepare(self) -> None:
        self.__done = False

    def __next__(self) -> T:
        if self.__done:
            raise StopIteration()

        obj = self._iterator.__next__()
        if not self.__predicate.Apply(obj):
            self.__done = True
            raise StopIteration()

        return obj


class _DropWhileIterable(_GenericIterable[T]):
    __slots__ = ("__predicate", "__done")

    def __init__(self, it: Iterable[T], predicate: Predicate[T]) -> None:
        super().__init__(it)
        self.__done = False
        self.__predicate = predicate

    def _prepare(self) -> None:
        self.__done = False

    def __next__(self) -> T:
        if self.__done:
            return self._iterator.__next__()
        while not self.__done:
            obj = self._iterator.__next__()
            if not self.__predicate.Apply(obj):
                self.__done = True
                return obj
        raise StopIteration()


class _ConcatIterable(_GenericIterable[T]):
    __slots__ = ("__iterable2", "__iterator2", "__done")

    def __init__(self, it1: Iterable[T], it2: Iterable[T]) -> None:
        super().__init__(it1)
        self.__done = False
        self.__iterable2 = it2
        self.__iterator2 = self.__iterable2.__iter__()

    def _prepare(self) -> None:
        self.__done = False
        self.__iterator2 = self.__iterable2.__iter__()

    def __next__(self) -> T:
        if self.__done:
            return self.__iterator2.__next__()
        try:
            return self._iterator.__next__()
        except StopIteration:
            self.__done = True
            return self.__next__()


class _DistinctIterable(_GenericIterable[T]):
    __slots__ = ("__set",)

    def __init__(self, it: Iterable[T]) -> None:
        super().__init__(it)
        self.__set: set[T] = set()

    def _prepare(self) -> None:
        self.__set = set()

    def __next__(self) -> T:
        obj = self._iterator.__next__()
        if obj not in self.__set:
            self.__set.add(obj)
            return obj
        return self.__next__()


class _MapIterable(Generic[T, V], Iterator[V], Iterable[V]):
    __slots__ = ("_iterable", "_iterator", "__mapper")

    def __init__(self, it: Iterable[T], mapper: Mapper[T, V]) -> None:
        self._iterable = it
        self._iterator = self._iterable.__iter__()
        self.__mapper = mapper

    def _prepare(self) -> None:
        pass

    def __iter__(self) -> Iterator[V]:
        self._iterator = self._iterable.__iter__()
        self._prepare()
        return self

    def __next__(self) -> V:
        return self.__mapper.Map(self._iterator.__next__())


class Stream(Generic[T]):
    __slots__ = ("__arg",)

    def __init__(self, arg: Iterable[T]) -> None:
        self.__arg = arg

    @staticmethod
    def of(arg: Iterable[T]) -> "Stream[T]":
        return Stream(arg)

    def map(self, mapper: Union[Mapper[T, V], Callable[[T], V]]) -> "Stream[V]":
        """
        Produces a new stream by mapping the stream elements using the given mapper function.
        Args:
            mapper (Union[Mapper[T, V], Callable[[T], V]]): The mapper

        Returns:
            Stream[V]: The result stream
        """
        return Stream(_MapIterable(self.__arg, mapperOf(mapper)))

    def flatMap(
        self, mapper: Union[Mapper[T, Iterable[V]], Callable[[T], Iterable[V]]]
    ) -> "Stream[V]":
        """
        Produces a flat stream by mapping an element of this stream to an iterable, then concatenates
        the iterables into a single stream.
        Args:
            mapper (Union[Mapper[T, Iterable[V]], Callable[[T], Iterable[V]]]): The mapper

        Returns:
            Stream[V]: the result stream
        """
        return Stream(flatMap(self.__arg, mapperOf(mapper)))

    def flatten(self, typ: type[V]) -> "Stream[V]":
        """
        Flattens a stream of iterables.
        CAUTION: This method will actually iterate the entire iterable, so if you're using
        infinite generators, calling this method will block the execution of the program.
        Returns:
            Stream[T]: A flattened stream
        """
        return self.flatMap(
            lambda v: cast(Iterable[V], v) if isinstance(v, Iterable) else [cast(V, v)]
        )

    def first(self) -> Opt[T]:
        """
        Finds and returns the first element of the stream.

        Returns:
            Opt[T]: First element
        """
        return self.findFirst(lambda e: True)

    def findFirst(self, predicate: Union[Predicate[T], Callable[[T], bool]]) -> Opt[T]:
        """
        Finds and returns the first element matching the predicate

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

        Returns:
            Opt[T]: The firs element found
        """
        return Opt(findFirst(self.__arg, predicateOf(predicate)))

    def filter(
        self, predicate: Union[Predicate[T], Callable[[T], bool]]
    ) -> "Stream[T]":
        """
        Returns a stream of objects that match the given predicate

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

        Returns:
            Stream[T]: The stream of filtered objects
        """

        return Stream(_FilterIterable(self.__arg, predicateOf(predicate)))

    def cast(self, castTo: type[V]) -> "Stream[V]":
        """
        Returns a stream of objects casted to the given type. Useful when receiving untyped data lists
        and they need to be used in a typed context.

        Args:
            castTo (type[V]): The type all objects will be casted to

        Returns:
            Stream[V]: The stream of casted objects
        """
        return Stream(_CastIterable(self.__arg, castTo))

    def anyMatch(self, predicate: Union[Predicate[T], Callable[[T], bool]]) -> bool:
        """
        Checks if any stream object matches the given predicate

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

        Returns:
            bool: True if any object matches, False otherwise
        """
        return self.filter(predicateOf(predicate)).isNotEmpty()

    def noneMatch(self, predicate: Union[Predicate[T], Callable[[T], bool]]) -> bool:
        """
        Checks if none of the stream objects matches the given predicate. This is the inverse of 'anyMatch`
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

        Returns:
            bool: True if no object matches, False otherwise
        """
        return self.filter(predicateOf(predicate)).isEmpty()

    def allMatch(self, predicate: Union[Predicate[T], Callable[[T], bool]]) -> bool:
        """
        Checks if all of the stream objects matche the given predicate.
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

        Returns:
            bool: True if all objects matche, False otherwise
        """
        return len(self.filter(predicateOf(predicate)).toList()) == len(
            list(self.__arg)
        )

    def isEmpty(self) -> bool:
        """
        Checks if the stream is empty
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Returns:
            bool: True if the stream is empty, False otherwise
        """
        return isEmptyOrNone(self.__arg)

    def isNotEmpty(self) -> bool:
        """
        Checks if the stream is not empty
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Returns:
            bool: True if the stream is not empty, False otherwise
        """
        return not isEmptyOrNone(self.__arg)

    def collect(self) -> Iterable[T]:
        """
        Returns an iterable with the content of the stream

        Returns:
            Iterable[T]: The iterable
        """
        return self.__arg

    def toList(self) -> list[T]:
        """
        Creates a list with the contents of the stream
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Returns:
            list[T]: The list
        """
        return list(self.__arg)

    def toSet(self) -> set[T]:
        """
        Creates a set with the contents of the stream
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Returns:
            set[T]: The set
        """
        return set(self.__arg)

    def toDict(
        self,
        keyMapper: Union[Mapper[T, V], Callable[[T], V]],
        valueMapper: Union[Mapper[T, K], Callable[[T], K]],
    ) -> dict[V, K]:
        """
        Creates a dictionary with the contents of the stream creating keys using
        the given key mapper and values using the value mapper
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Args:
            keyMapper (Union[Mapper[T, V], Callable[[T], V]]): The key mapper
            valueMapper (Union[Mapper[T, K], Callable[[T], K]]): The value mapper

        Returns:
            dict[V, K]: The resulting dictionary
        """
        keyMapperObj = mapperOf(keyMapper)
        valueMapperObj = mapperOf(valueMapper)
        return {keyMapperObj.Map(v): valueMapperObj.Map(v) for v in self.__arg}

    def toDictAsValues(
        self, keyMapper: Union[Mapper[T, V], Callable[[T], V]]
    ) -> dict[V, T]:
        """
        Creates a dictionary with the contents of the stream creating keys using
        the given key mapper
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Args:
            keyMapper (Union[Mapper[T, V], Callable[[T], V]]): The key mapper

        Returns:
            dict[V, T]: The resulting dictionary
        """
        keyMapperObj = mapperOf(keyMapper)
        return {keyMapperObj.Map(v): v for v in self.__arg}

    def toDictAsKeys(
        self, valueMapper: Union[Mapper[T, V], Callable[[T], V]]
    ) -> dict[T, V]:
        """
        Creates a dictionary using the contents of the stream as keys and mapping
        the dictionary values using the given value mapper
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Args:
            keyMapper (Union[Mapper[T, V], Callable[[T], V]]): The value mapper

        Returns:
            dict[V, T]: The resulting dictionary
        """
        valueMapperObj = mapperOf(valueMapper)
        return {v: valueMapperObj.Map(v) for v in self.__arg}

    def each(self, action: Callable[[T], Any]) -> "Stream[T]":
        """
        Executes the action callable for each of the stream's elements.
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Args:
            action (Callable[[T], Any]): The action
        """
        each(self.__arg, action)
        return self

    def ofType(self, theType: type[V]) -> "Stream[V]":
        """
        Returns all items of the given type as a stream

        Args:
            theType (type[V]): The given type

        Returns:
            Stream[V]: The result stream
        """
        return self.filter(ClassOps(theType).instanceOf).cast(theType)

    def skip(self, count: int) -> "Stream[T]":
        """
        Returns a stream without the first number of items specified by 'count'

        Args:
            count (int): How many items should be skipped

        Returns:
            Stream[T]: The result stream
        """
        return Stream(_SkipIterable(self.__arg, count))

    def limit(self, count: int) -> "Stream[T]":
        """
        Returns a stream limited to the first 'count' items of this stream

        Args:
            count (int): The max amount of items

        Returns:
            Stream[T]: The result stream
        """
        return Stream(_LimitIterable(self.__arg, count))

    def takeWhile(
        self, predicate: Union[Predicate[T], Callable[[T], bool]]
    ) -> "Stream[T]":
        """
        Returns a stream of elements until the first element that DOES NOT match the given predicate

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

        Returns:
            Stream[T]: The result stream
        """
        return Stream(_TakeWhileIterable(self.__arg, predicateOf(predicate)))

    def dropWhile(
        self, predicate: Union[Predicate[T], Callable[[T], bool]]
    ) -> "Stream[T]":
        """
        Returns a stream of elements by dropping the first elements that match the given predicate

        Args:
            predicate (Union[Predicate[T], Callable[[T], bool]]): The predicate

        Returns:
            Stream[T]: The result stream
        """
        return Stream(_DropWhileIterable(self.__arg, predicateOf(predicate)))

    def reduce(self, reducer: Union[Reducer[T], Callable[[T, T], T]]) -> Opt[T]:
        """
        Reduces a stream to a single value. The reducer takes two values and
        returns only one. This function can be used to find min or max from a stream of ints.
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Args:
            reducer (Union[Reducer[T], Callable[[T, T], T]]): The reducer

        Returns:
            Opt[T]: The resulting optional
        """
        return Opt(reduce(self.__arg, reducer))

    def nonNull(self) -> "Stream[T]":
        """
        Returns a stream of non null objects from this stream

        Returns:
            Stream[T]: The result stream
        """
        return self.filter(isNotNone)

    def sort(self, comparator: Callable[[T, T], int]) -> "Stream[T]":
        """
        Returns a stream with the elements sorted according to the comparator function.
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Args:
            comparator (Callable[[T, T], int]): The comparator function

        Returns:
            Stream[T]: The resulting stream
        """
        return Stream(sort(list(self.__arg), comparator))

    def reverse(self) -> "Stream[T]":
        """
        Returns a the reverted stream.
        CAUTION: This method will actually iterate the entire stream, so if you're using
        infinite generators, calling this method will block the execution of the program.

        Returns:
            Stream[T]: Thje resulting stream
        """
        elems = list(self.__arg)
        elems.reverse()
        return Stream(elems)

    def distinct(self) -> "Stream[T]":
        """
        Returns disting elements from the stream.
        CAUTION: Use this method on stream of items that have the __eq__ method implemented,
        otherwise the method will consider all values distinct

        Returns:
            Stream[T]: The resulting stream
        """
        if self.__arg is None:
            return self
        return Stream(_DistinctIterable(self.__arg))

    def concat(self, newStream: "Stream[T]") -> "Stream[T]":
        """
        Returns a stream concatenating the values from this stream with the ones
        from the given stream.

        Args:
            newStream (Stream[T]): The stream to be concatenated with

        Returns:
            Stream[T]: The resulting stream
        """
        return Stream(_ConcatIterable(self.__arg, newStream.__arg))


def stream(it: Iterable[T]) -> Stream[T]:
    """
    Helper method, equivalent to Stream(it)

    Args:
        it (Iterable[T]): The iterator

    Returns:
        Stream[T]: The stream
    """
    return Stream(it)


def optional(val: Optional[T]) -> Opt[T]:
    """
    Helper method, equivalent to Opt(val)

    Args:
        val (Optional[T]): The value

    Returns:
        Opt[T]: The optional
    """
    return Opt(val)



from baseTest import BaseTestCase
from jstreams.predicate import Not, contains, isInInterval, isZero
from jstreams.tuples import leftMatches, middleMatches, pair, rightMatches, triplet


class TestTuples(BaseTestCase):
    def test_pair(self) -> None:
        v = pair("a", 0)
        self.assertEqual(v.left(), "a", "Left should be correct")
        self.assertEqual(v.right(), 0, "Right should be correct")

    def test_triplet(self) -> None:
        v = triplet("test", 1, None)
        self.assertEqual(v.left(), "test", "Left should be correct")
        self.assertEqual(v.middle(), 1, "Middle should be correct")
        self.assertIsNone(v.right(), "Right should be None")

    def test_pair_predicate(self) -> None:
        v = pair("test", 0)
        self.assertTrue(leftMatches(contains("es"))(v), "Left should match predicate")
        self.assertFalse(
            leftMatches(contains("as"))(v), "Left should not match predicate"
        )
        self.assertTrue(rightMatches(isZero)(v), "Right should match predicate")
        self.assertFalse(
            rightMatches(Not(isZero))(v), "Right should not match predicate"
        )

    def test_triplet_predicate(self) -> None:
        v = triplet("test", 0, 1.5)
        self.assertTrue(leftMatches(contains("es"))(v), "Left should match predicate")
        self.assertFalse(
            leftMatches(contains("as"))(v), "Left should not match predicate"
        )
        self.assertTrue(middleMatches(isZero)(v), "Middle should match predicate")
        self.assertFalse(
            middleMatches(Not(isZero))(v), "Middle should not match predicate"
        )
        self.assertTrue(
            rightMatches(isInInterval(1, 2))(v), "Right should match predicate"
        )
        self.assertFalse(
            rightMatches(isInInterval(1.6, 2))(v), "Right should not match predicate"
        )



from threading import Lock
from typing import Any, Callable, Generic, Optional, TypeAlias, TypeVar, Union, cast
from jstreams.noop import NoOp, NoOpCls

AnyDict: TypeAlias = dict[str, Any]


class AutoStart:
    __slots__ = ()
    """
    Interface notifying the container that a component must be started as soon as it
    is added to the container.
    """

    def start(self) -> None:
        pass


class AutoInit:
    __slots__ = ()
    """
    Interface notifying the container that a component must be initialized by calling the 'init' method
    as soon as it is added to the container.
    """

    def init(self) -> None:
        pass


class ContainerDependency:
    def __init__(self) -> None:
        self.qualifiedDependencies: AnyDict = {}


class VariableDependency:
    def __init__(self) -> None:
        self.qualifiedVariables: AnyDict = {}


T = TypeVar("T")


class _Injector:
    instance: Optional["_Injector"] = None
    instanceLock: Lock = Lock()

    def __init__(self) -> None:
        self.__components: dict[type, ContainerDependency] = {}
        self.__variables: dict[type, VariableDependency] = {}

    def clear(self) -> None:
        self.__components = {}
        self.__variables = {}

    def get(self, className: type[T], qualifier: Optional[str] = None) -> T:
        if (foundObj := self.find(className, qualifier)) is None:
            raise ValueError("No object found for class " + str(className))
        return foundObj

    def getVar(self, className: type[T], qualifier: str) -> T:
        if (foundVar := self.findVar(className, qualifier)) is None:
            raise ValueError(
                "No variable found for class "
                + str(className)
                + " and qualifier "
                + qualifier
            )
        return foundVar

    def findVar(self, className: type[T], qualifier: str) -> Optional[T]:
        foundVar = self._getVar(className, qualifier)
        return foundVar if foundVar is None else cast(T, foundVar)

    def findVarOr(self, className: type[T], qualifier: str, orVal: T) -> Optional[T]:
        foundVar = self._getVar(className, qualifier)
        return orVal if foundVar is None else cast(T, foundVar)

    def find(self, className: type[T], qualifier: Optional[str] = None) -> Optional[T]:
        foundObj = self._get(className, qualifier)
        return foundObj if foundObj is None else cast(T, foundObj)

    def findOr(
        self,
        className: type[T],
        orCall: Callable[[], T],
        qualifier: Optional[str] = None,
    ) -> T:
        foundObj = self._get(className, qualifier)
        return orCall() if foundObj is None else cast(T, foundObj)

    def findNoOp(
        self, className: type[T], qualifier: Optional[str] = None
    ) -> Union[T, NoOpCls]:
        if (foundObj := self.find(className, qualifier)) is None:
            return NoOp
        return foundObj

    @staticmethod
    def getInstance() -> "_Injector":
        # If the instance is not initialized
        if _Injector.instance is None:
            # Lock for instantiation
            with _Injector.instanceLock:
                # Check if the instance was not already initialized before acquiring the lock
                if _Injector.instance is None:
                    # Initialize
                    _Injector.instance = _Injector()
        return _Injector.instance

    def provideVarIfNotNull(
        self, className: type, qualifier: str, value: Any
    ) -> "_Injector":
        if value is not None:
            self.provideVar(className, qualifier, value)
        return self

    def provideVar(self, className: type, qualifier: str, value: Any) -> "_Injector":
        if (varDep := self.__variables.get(className)) is None:
            varDep = VariableDependency()
            self.__variables[className] = varDep
        if qualifier is None:
            qualifier = ""
        varDep.qualifiedVariables[qualifier] = value
        return self

    # Register a component with the container
    def provide(
        self, className: type, comp: Any, qualifier: Optional[str] = None
    ) -> "_Injector":
        if (containerDep := self.__components.get(className)) is None:
            containerDep = ContainerDependency()
            self.__components[className] = containerDep
        if qualifier is None:
            qualifier = ""
        containerDep.qualifiedDependencies[qualifier] = comp
        return self

    # Get a component from the container
    def _get(self, className: type, qualifier: Optional[str]) -> Any:
        if (containerDep := self.__components.get(className)) is None:
            return None
        if qualifier is None:
            qualifier = ""
        return containerDep.qualifiedDependencies.get(qualifier, None)

    def _getVar(self, className: type, qualifier: str) -> Any:
        if (varDep := self.__variables.get(className)) is None:
            return None
        return varDep.qualifiedVariables.get(qualifier, None)

    def provideDependencies(self, dependencies: dict[type, Any]) -> "_Injector":
        for componentClass in dependencies:
            service = dependencies[componentClass]
            self.provide(componentClass, service)
            if isinstance(service, AutoInit):
                service.init()
            if isinstance(service, AutoStart):
                service.start()
        return self

    def provideVariables(self, variables: list[tuple[type, str, Any]]) -> "_Injector":
        for varClass, qualifier, value in variables:
            self.provideVar(varClass, qualifier, value)
        return self


Injector = _Injector.getInstance()


def injector() -> _Injector:
    return Injector


def inject(className: type[T], qualifier: Optional[str] = None) -> T:
    return injector().get(className, qualifier)


def var(className: type[T], qualifier: str) -> T:
    return injector().getVar(className, qualifier)


def resolveDependencies(dependencies: dict[str, type]) -> Callable[[type[T]], type[T]]:
    def wrap(cls: type[T]) -> type[T]:
        for key, typ in dependencies.items():
            if key.startswith("__"):
                raise ValueError(
                    "Cannot inject private attribute. Only public and protected attributes can use injection"
                )
            setattr(cls, key, injector().find(typ))
        return cls

    return wrap


class InjectedDependency(Generic[T]):
    __slots__ = ["__typ"]

    def __init__(self, typ: type[T]) -> None:
        self.__typ = typ

    def get(self) -> T:
        return injector().get(self.__typ)


class OptionalInjectedDependency(Generic[T]):
    __slots__ = ["__typ"]

    def __init__(self, typ: type[T]) -> None:
        self.__typ = typ

    def get(self) -> Optional[T]:
        return injector().find(self.__typ)



from typing import Any
from baseTest import BaseTestCase
from jstreams import (
    BehaviorSubject,
    Flowable,
    PublishSubject,
    ReplaySubject,
    Single,
    rxMap,
    rxFilter,
    rxReduce,
    rxTake,
    rxTakeWhile,
    rxTakeUntil,
    rxDropWhile,
    rxDropUntil,
    rxDrop,
)


class TestRx(BaseTestCase):
    def __init__(self, methodName="runTest"):
        super().__init__(methodName)
        self.val = None
        self.val2 = None

    def setVal(self, val: Any) -> None:
        self.val = val

    def addVal(self, val: Any) -> None:
        self.val.append(val)

    def addVal2(self, val: Any) -> None:
        self.val2.append(val)

    def test_single(self) -> None:
        self.val = None
        Single("test").subscribe(self.setVal)
        self.assertEqual(self.val, "test")

    def test_flowable(self) -> None:
        self.val = []
        init = ["test1", "test2"]
        Flowable(init).subscribe(self.addVal)
        self.assertListEqual(init, self.val)

    def test_behavior_subject(self) -> None:
        subject = BehaviorSubject("1")
        self.val = []
        sub = subject.subscribe(self.addVal)
        self.assertListEqual(
            self.val,
            ["1"],
            "BehaviorSubject should push the latest value on subscription",
        )
        subject.onNext("2")
        self.assertListEqual(
            self.val,
            ["1", "2"],
            "BehaviorSubject should push the latest value after subscription",
        )
        subject.onNext("3")
        self.assertListEqual(
            self.val,
            ["1", "2", "3"],
            "BehaviorSubject should push the latest value after subscription",
        )
        subject.pause(sub)
        subject.onNext("4")
        self.assertListEqual(
            self.val,
            ["1", "2", "3"],
            "BehaviorSubject should not push the latest value when subscription is paused",
        )
        subject.resumePaused()
        subject.onNext("5")
        self.assertListEqual(
            self.val,
            ["1", "2", "3", "5"],
            "BehaviorSubject should push the latest value when subscription is resumed",
        )
        subject.dispose()

    def test_publish_subject(self) -> None:
        subject = PublishSubject(str)
        self.val = []
        subject.onNext("1")
        sub = subject.subscribe(self.addVal)
        self.assertListEqual(
            self.val,
            [],
            "PublishSubject should not push the latest value on subscription",
        )
        subject.onNext("2")
        self.assertListEqual(
            self.val,
            ["2"],
            "PublishSubject should push the latest value after subscription",
        )
        subject.onNext("3")
        self.assertListEqual(
            self.val,
            ["2", "3"],
            "PublishSubject should push the latest value after subscription",
        )
        subject.pause(sub)
        subject.onNext("4")
        self.assertListEqual(
            self.val,
            ["2", "3"],
            "PublishSubject should not push the latest value when subscription is paused",
        )
        subject.resumePaused()
        subject.onNext("5")
        self.assertListEqual(
            self.val,
            ["2", "3", "5"],
            "PublishSubject should push the latest value when subscription is resumed",
        )
        subject.dispose()

    def test_replay_subject(self) -> None:
        subject = ReplaySubject(["A", "B", "C"])
        self.val = []
        self.val2 = []
        sub1 = subject.subscribe(self.addVal)
        self.assertListEqual(self.val, ["A", "B", "C"])
        subject.onNext("1")
        self.assertListEqual(self.val, ["A", "B", "C", "1"])
        sub2 = subject.subscribe(self.addVal2)
        self.assertListEqual(self.val2, ["A", "B", "C", "1"])
        subject.dispose()

    def test_replay_subject_map(self) -> None:
        subject = ReplaySubject(["a1", "a2", "a3"])
        self.val = []
        subject.pipe(rxMap(str.upper)).subscribe(self.addVal)
        self.assertListEqual(self.val, ["A1", "A2", "A3"])
        subject.onNext("a4")
        self.assertListEqual(self.val, ["A1", "A2", "A3", "A4"])
        subject.dispose()

    def test_replay_subject_filter(self) -> None:
        subject = ReplaySubject(["a1", "a2", "a3", "b", "c", "a4"])
        self.val = []
        subject.pipe(rxFilter(lambda s: s.startswith("a"))).subscribe(self.addVal)
        self.assertListEqual(self.val, ["a1", "a2", "a3", "a4"])
        subject.onNext("a5")
        self.assertListEqual(self.val, ["a1", "a2", "a3", "a4", "a5"])
        subject.onNext("b")
        self.assertListEqual(self.val, ["a1", "a2", "a3", "a4", "a5"])
        subject.dispose()

    def test_replay_subject_map_and_filter(self) -> None:
        subject = ReplaySubject(["a1", "a2", "a3"])
        self.val = []
        pipe1 = subject.pipe(rxMap(str.upper), rxFilter(lambda s: s.endswith("3")))
        pipe1.subscribe(self.addVal)
        self.assertListEqual(self.val, ["A3"])
        subject.onNext("a4")
        self.assertListEqual(self.val, ["A3"])
        subject.dispose()

    def test_replay_subject_map_and_filter_multiple(self) -> None:
        subject = ReplaySubject(["a1", "a2", "a3"])
        self.val = []
        pipe1 = subject.pipe(
            rxMap(str.upper),
            rxFilter(lambda s: s.endswith("3")),
            rxMap(lambda s: s + "Test"),
        )
        pipe1.subscribe(self.addVal)
        self.assertListEqual(self.val, ["A3Test"])
        subject.onNext("a4")
        self.assertListEqual(self.val, ["A3Test"])
        subject.dispose()

    def test_replay_subject_filter_and_reduce(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        self.val = []
        pipe1 = subject.pipe(
            rxFilter(lambda nr: nr <= 10), rxReduce(lambda a, b: max(a, b))
        )
        pipe1.subscribe(self.addVal)
        self.assertListEqual(self.val, [1, 7])
        subject.onNext(9)
        self.assertListEqual(self.val, [1, 7, 9])
        subject.dispose()

    def test_replay_subject_with_take(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        self.val = []
        pipe1 = subject.pipe(rxTake(int, 3))
        pipe1.subscribe(self.addVal)
        self.assertListEqual(self.val, [1, 7, 20])
        subject.onNext(9)
        self.assertListEqual(self.val, [1, 7, 20])
        subject.dispose()

    def test_replay_subject_with_takeWhile(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        self.val = []
        pipe1 = subject.pipe(rxTakeWhile(lambda v: v < 10))
        pipe1.subscribe(self.addVal)
        self.assertListEqual(self.val, [1, 7])
        subject.onNext(9)
        self.assertListEqual(self.val, [1, 7])
        subject.dispose()

    def test_replay_subject_with_takeUntil(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        self.val = []
        pipe1 = subject.pipe(rxTakeUntil(lambda v: v > 10))
        pipe1.subscribe(self.addVal)
        self.assertListEqual(self.val, [1, 7])
        subject.onNext(9)
        self.assertListEqual(self.val, [1, 7])
        subject.dispose()

    def test_replay_subject_with_drop(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        self.val = []
        pipe1 = subject.pipe(rxDrop(int, 3))
        pipe1.subscribe(self.addVal)
        self.assertListEqual(self.val, [5, 100, 40])
        subject.onNext(9)
        self.assertListEqual(self.val, [5, 100, 40, 9])
        subject.dispose()

    def test_replay_subject_with_dropWhile(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        self.val = []
        pipe1 = subject.pipe(rxDropWhile(lambda v: v < 100))
        pipe1.subscribe(self.addVal)
        self.assertListEqual(self.val, [100, 40])
        subject.onNext(9)
        self.assertListEqual(self.val, [100, 40, 9])
        subject.dispose()

    def test_replay_subject_with_dropUntil(self) -> None:
        subject = ReplaySubject([1, 7, 20, 5, 100, 40])
        self.val = []
        pipe1 = subject.pipe(rxDropUntil(lambda v: v > 20))
        pipe1.subscribe(self.addVal)
        self.assertListEqual(self.val, [100, 40])
        subject.onNext(9)
        self.assertListEqual(self.val, [100, 40, 9])
        subject.dispose()

    def test_replay_subject_with_pipe_chaining(self) -> None:
        subject = ReplaySubject(range(1, 100))
        self.val = []
        self.val2 = []
        chainedPipe = (
            subject.pipe(rxTakeUntil(lambda e: e > 20))
            .pipe(rxFilter(lambda e: e % 2 == 0))
            .pipe(rxTakeWhile(lambda e: e < 10))
        )
        chainedPipe.subscribe(self.addVal)
        chainedPipe.subscribe(self.addVal2)
        chainedPipe.dispose()
        self.assertListEqual(self.val, [2, 4, 6, 8])
        self.assertListEqual(self.val2, [2, 4, 6, 8])



from typing import Any


class NoOpCls:
    """No-op that can be put in lieu of almost anything.
    Any actions on that object will be just ignored. You can
    call it, get its atributes, use a a context, anything :)
    Cool feature
    ------------
    NoOp tracks its levels of absurdity.
    Examples
    --------
    >>> from no_op import NoOp as xyz
    # Now all calls to xyz are simply ignored without any exception:
    >>>  print(xyz.get('qwerty').wtf['dupa'].callme('tesla', model='X')[5, ...])
    Just no-op on level 7.

    Copyright
    ---------
    (c) 2021 Maciej J. Mikulski
    Feel free to use it, I put NoOp in public domain.
    """

    def __init__(self, level: int = 0) -> None:
        self.level = level

    def __getattr__(self, attr: Any) -> Any:
        return NoOpCls(self.level + 1)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return NoOpCls(self.level + 1)

    def __getitem__(self, item: Any) -> Any:
        return NoOpCls(self.level + 1)

    def __enter__(self) -> Any:
        return NoOpCls(self.level + 1)

    def __repr__(self) -> str:
        return f"Just no-op on level {self.level}."


NoOp = NoOpCls()


def noop() -> NoOpCls:
    return NoOp
